from dexy.doc import Doc
from tests.utils import tempdir
from tests.utils import wrap
from dexy.wrapper import Wrapper
import os
import tarfile
import zipfile

def test_zip_archive_filter():
    with tempdir():
        with open("hello.py", "w") as f:
            f.write("print 'hello'")

        with open("hello.rb", "w") as f:
            f.write("puts 'hello'")

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper = Wrapper()

        doc = Doc("archive.zip|zip",
                wrapper,
                [
                    Doc("hello.py", wrapper),
                    Doc("hello.rb", wrapper),
                    Doc("hello.py|pyg", wrapper),
                    Doc("hello.rb|pyg", wrapper)
                    ],
                contents=" ")

        wrapper.run_docs(doc)
        wrapper.report()

        path_exists = os.path.exists("output/archive.zip")
        assert path_exists
        z = zipfile.ZipFile("output/archive.zip", "r")
        names = z.namelist()
        assert "archive/hello.py" in names
        assert "archive/hello.rb" in names
        assert "archive/hello.py-pyg.html" in names
        assert "archive/hello.rb-pyg.html" in names
        z.close()

def test_archive_filter():
    with wrap() as wrapper:
        with open("hello.py", "w") as f:
            f.write("print 'hello'")

        with open("hello.rb", "w") as f:
            f.write("puts 'hello'")

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper = Wrapper()

        doc = Doc("archive.tgz|archive",
                wrapper,
                [
                    Doc("hello.py", wrapper),
                    Doc("hello.rb", wrapper),
                    Doc("hello.py|pyg", wrapper),
                    Doc("hello.rb|pyg", wrapper)
                ],
                contents=" ")

        wrapper.run_docs(doc)
        wrapper.report()

        assert os.path.exists("output/archive.tgz")
        tar = tarfile.open("output/archive.tgz", mode="r:gz")
        names = tar.getnames()
        assert "archive/hello.py" in names
        assert "archive/hello.rb" in names
        assert "archive/hello.py-pyg.html" in names
        assert "archive/hello.rb-pyg.html" in names
        tar.close()

def test_archive_filter_with_short_names():
    with wrap() as wrapper:
        with open("hello.py", "w") as f:
            f.write("print 'hello'")

        with open("hello.rb", "w") as f:
            f.write("puts 'hello'")

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper = Wrapper()

        doc = Doc("archive.tgz|archive",
                wrapper,
                [
                    Doc("hello.py", wrapper),
                    Doc("hello.rb", wrapper),
                    Doc("hello.py|pyg", wrapper),
                    Doc("hello.rb|pyg", wrapper)
                    ],
                contents=" ",
                archive={'use-short-names' : True}
                )

        wrapper.run_docs(doc)
        wrapper.report()


        assert os.path.exists("output/archive.tgz")
        tar = tarfile.open("output/archive.tgz", mode="r:gz")
        names = tar.getnames()
        assert "archive/hello.py" in names
        assert "archive/hello.rb" in names
        assert "archive/hello.py.html" in names
        assert "archive/hello.rb.html" in names
        tar.close()

def test_unprocessed_directory_archive_filter():
    with wrap() as wrapper:
        with open("abc.txt", "w") as f:
            f.write('this is abc')

        with open("def.txt", "w") as f:
            f.write('this is def')

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper = Wrapper()

        doc = Doc("archive.tgz|tgzdir",
                wrapper,
                [],
                contents="ignore",
                tgzdir={'dir' : '.'}
                )
        wrapper.run_docs(doc)
        wrapper.report()

        assert os.path.exists("output/archive.tgz")
        tar = tarfile.open("output/archive.tgz", mode="r:gz")
        names = tar.getnames()

        assert ("./abc.txt" in names) or ("abc.txt" in names)
        assert ("./def.txt" in names) or ("def.txt" in names)
        tar.close()



from dexy.reporter import Reporter

class Graphviz(Reporter):
    """
    Emits a graphviz graph of the network structure.
    """
    aliases = ['graphviz', 'nodegraph']
    _settings = {
            'in-cache-dir' : True,
            'filename' : 'graph.dot',
            "run-for-wrapper-states" : ["ran", "checked", "error"]
            }

    def run(self, wrapper):
        self.wrapper = wrapper
        def print_children(node, indent=0):
            content = []
            content.append(node.key)
            for child in node.children:
                for line in print_children(child, indent+1):
                    content.append(line)
            return content

        def print_inputs(node):
            content = []
            for child in node.inputs:
                content.extend(print_inputs(child))
                content.append("\"%s\" -> \"%s\";" % (node, child))
            return content

        graph = []
        graph.append("digraph G {")
        for node in list(wrapper.nodes.values()):
            graph.extend(print_inputs(node))
        graph.append("}")

        self.create_cache_reports_dir()
        with open(self.report_file(), "w") as f:
            f.write("\n".join(graph))




# Extra Pygments Lexers - will be submitted to Pygments and removed from here
# when available in a Pygments release

from pygments.lexers.templates import DjangoLexer
from pygments.lexer import DelegatingLexer
from pygments.lexers.text import RstLexer

class RstDjangoLexer(DelegatingLexer):
    """
    Subclass of the `DjangoLexer` that highlights unlexed data with the
    `RstLexer`.
    """

    name = 'ReStructuredText+Django/Jinja'
    aliases = ['rst+django', 'rst+jinja']
    filenames = ['*.rst']

    def __init__(self, **options):
        super(RstDjangoLexer, self).__init__(RstLexer, DjangoLexer, **options)

    def analyse_text(text):
        rv = DjangoLexer.analyse_text(text) - 0.01
        # TODO count how many times ".. " is in text
        if ".. " in text:
            rv += 0.4
        return rv



from dexy.reporter import Reporter
import os

class Output(Reporter):
    """
    Creates canonical dexy output with files given short filenames.
    """
    aliases = ['output']
    _settings = {
            'dir' : 'output'
            }

    def write_canonical_data(self, doc):
        output_name = doc.output_data().output_name()

        if output_name:
            fp = os.path.join(self.setting('dir'), output_name)

            if fp in self.locations:
                self.log_warn("WARNING overwriting file %s" % fp)
            else:
                self.locations[fp] = []
            self.locations[fp].append(doc.key)

            parent_dir = os.path.dirname(fp)
            try:
                os.makedirs(parent_dir)
            except os.error:
                pass

            self.log_debug("  writing %s to %s" % (doc.key, fp))

            doc.output_data().output_to_file(fp)

    def run(self, wrapper):
        self.wrapper=wrapper
        self.locations = {}

        self.remove_reports_dir(self.wrapper, keep_empty_dir=True)
        self.create_reports_dir()
        for doc in list(wrapper.nodes.values()):
            if not doc.key_with_class() in wrapper.batch.docs:
                continue
            if not doc.state in ('ran', 'consolidated'):
                continue
            if not hasattr(doc, 'output_data'):
                continue

            if doc.output_data().is_canonical_output():
                self.write_canonical_data(doc)

class LongOutput(Reporter):
    """
    Creates complete dexy output with files given long, unique filenames.
    """
    aliases = ['long']
    _settings = {
            'default' : False,
            'dir' : 'output-long'
            }

    def run(self, wrapper):
        self.wrapper=wrapper
        self.create_reports_dir()
        for doc in list(wrapper.nodes.values()):
            if not doc.key_with_class() in wrapper.batch.docs:
                continue
            if not doc.state in ('ran', 'consolidated'):
                continue
            if not hasattr(doc, 'output_data'):
                continue

            fp = os.path.join(self.setting('dir'), doc.output_data().long_name())

            try:
                os.makedirs(os.path.dirname(fp))
            except os.error:
                pass

            self.log_debug("  writing %s to %s" % (doc.key, fp))
            doc.output_data().output_to_file(fp)



from tests.utils import wrap
from dexy.doc import Doc

python_file_content = """
import math

# Comment for foo
def foo():
    '''
    docstring for foo
    '''
    return True


# Comment for bar
def bar():
    '''
    docstring for bar
    '''
    return False


# Comment before decorator
@decorator
@another
def decorated():
    pass

@decorator(some, args)
def decorated_with_args():
    pass


class Foo(object):
    @decorator
    @another
    def decorated(self):
        pass


@decorator
class Decorated(object):
    pass
"""

def test_pyparse_filter_on_python_files():
    with wrap() as wrapper:
        doc = Doc("source.py|pyparse", wrapper, [], contents=python_file_content)
        wrapper.run_docs(doc)

        data = doc.output_data()
        keys = list(data.keys())

        assert 'bar:source' in keys
        assert 'foo:source' in keys
        
        assert data['foo:doc'] == "docstring for foo"
        assert data['foo:source'].startswith("def foo():\n")

        assert data['bar:doc'] == "docstring for bar"
        assert data['bar:source'].startswith("def bar():\n")

        assert data['decorated:source'].startswith("@decorator\n@another\ndef decorated():\n")

        assert data['Foo:source'].startswith("class Foo(object):\n")

        assert data['Foo.decorated:source'].startswith("    @decorator\n    @another\n    def decorated(self):\n")



from dexy.node import Node
from dexy.commands.utils import dummy_wrapper
import inspect

def nodes_command(
        alias = False # Print docs for a particular node type.
        ):
    """
    Prints available node types and their settings.
    """

    if not alias:
        # list all plugins
        for alias in sorted(Node.plugins):
            print(alias)

        print("For info on a particular node type run `dexy nodes -alias doc`")
    else:
        print_node_info(alias)


def print_node_info(alias):
    print(alias)

    _, settings = Node.plugins[alias]

    instance = Node.create_instance(alias, "dummy", dummy_wrapper())
    instance.update_settings(settings)

    print('')
    print(instance.setting('help'))
    print('')

    if len(instance._instance_settings) > 2:
        print('Settings:')

    for k in sorted(instance._instance_settings):
        if k in ('aliases', 'help',):
            continue

        tup = instance._instance_settings[k]
        print("    %s" % k)

        for line in inspect.cleandoc(tup[0]).splitlines():
            print("        %s" % line)

        print("        default value: %s" % tup[1])
        print('')



from dexy.filter import DexyFilter
import dexy.exceptions

class Deprecated(DexyFilter):
    """
    Base class for deprecated filters.
    """
    aliases = []

    def process(self):
        msg = "%s\n%s" % (self.artifact.key, self.__doc__)
        raise dexy.exceptions.UserFeedback(msg)

class FilenameFilter(Deprecated):
    """
    Deprecated. No longer needed.
    
    Dexy should now automatically detect new files that are created by your
    scripts if the add-new-files setting is true (which it is by default in
    many filters). You should remove '|fn' from your config and anywhere
    documents are referenced, and remove the 'dexy--' prefix from filenames in
    your scripts.
    """
    aliases = ['fn']



from dexy.commands.utils import print_indented
from dexy.commands.utils import print_rewrapped
import dexy.reporter

def reporters_command(
        alias=False, # Print detailed information about the specified reporter.
        simple=False, # Only print report aliases, without other information.
        ):
    """
    List available reports which dexy can run.
    """
    if simple:
        for reporter in dexy.reporter.Reporter:
            print(reporter.alias)

    elif alias:
        nodoc_settings = ('aliases', 'help',)

        reporter = dexy.reporter.Reporter.create_instance(alias)

        print_indented("%s Reporter" % reporter.__class__.__name__)
        print('')

        print_indented("settings:")
        print('')
        for name in sorted(reporter._instance_settings):
            if name in nodoc_settings:
                continue

            docs, default_value = reporter._instance_settings[name]
            print_indented(name, 2)
            print_rewrapped(docs, 4)
            print_indented("(default: %r)" % default_value, 4)
            print('')

        reporter.help()

        print('')

    else:
        FMT = "%-15s %-9s %s"
    
        print(FMT % ('alias', 'default', 'info'))
        for reporter in dexy.reporter.Reporter:
            help_text = reporter.setting('help').splitlines()[0]
            default_text = reporter.setting('default') and 'true' or 'false'
            print(FMT % (reporter.alias, default_text, help_text))



from io import StringIO
from dexy.data import Sectioned
from dexy.doc import Doc
from dexy.exceptions import InactivePlugin
from dexy.utils import char_diff
from dexy.utils import tempdir
from mock import MagicMock
from nose.exc import SkipTest
import os
import re
import sys

import dexy.load_plugins

TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')

def make_wrapper():
    from dexy.wrapper import Wrapper
    return Wrapper(log_level = 'DEBUG', debug=True)

class wrap(tempdir):
    """
    Create a temporary directory and initialize a dexy wrapper.
    """
    def __enter__(self):
        self.make_temp_dir()
        wrapper = make_wrapper()
        wrapper.create_dexy_dirs()
        wrapper = make_wrapper()
        wrapper.to_valid()
        return wrapper

    def __exit__(self, type, value, traceback):
        self.remove_temp_dir()
        if isinstance(value, InactivePlugin):
            print(value.message)
            raise SkipTest
            return True # swallow InactivePlugin error

class runfilter(wrap):
    """
    Create a temporary directory, initialize a doc and a wrapper, and run the doc.
    """
    def __init__(self, filter_alias, doc_contents, ext=".txt", basename=None):
        self.filter_alias = filter_alias
        self.doc_contents = doc_contents
        self.ext = ext
        self.basename = basename

    def __enter__(self):
        self.make_temp_dir()

        def run_example(doc_key, doc_contents):
            wrapper = make_wrapper()

            if isinstance(doc_contents, str):
                data_class = 'generic'
            else:
                data_class = 'sectioned'

            doc = Doc(
                    doc_key,
                    wrapper,
                    [],
                    contents = doc_contents,
                    data_class = data_class
                    )
            wrapper.run_docs(doc)
            return doc

        try:
            wrapper = make_wrapper()
            wrapper.create_dexy_dirs()

            if self.basename:
                filename = "%s%s" % (self.basename, self.ext)
            else:
                filename = "example%s" % self.ext

            doc_key = "subdir/%s|%s" % (filename, self.filter_alias)

            doc = run_example(doc_key, self.doc_contents)

        except InactivePlugin:
            raise SkipTest

        return doc

def assert_output(filter_alias, doc_contents, expected_output, ext=".txt", basename=None):
    if not ext.startswith("."):
        raise Exception("ext arg to assert_in_output must start with dot")

    if isinstance(doc_contents, dict):
        raise Exception("doc contents can't be dict")

    with runfilter(filter_alias, doc_contents, ext=ext, basename=basename) as doc:
        actual_output_data = doc.output_data()
        if isinstance(actual_output_data, Sectioned):
            for section_name, expected_section_contents in expected_output.items():
                try:
                    actual_section_contents = str(actual_output_data[section_name])
                    assert actual_section_contents == expected_section_contents
                except AssertionError:
                    print("Sections %s are not the same" % section_name)
                    print(char_diff(actual_section_contents, expected_section_contents))
        else:
            actual_output_data = str(doc.output_data())
            try:
                assert actual_output_data == expected_output
            except AssertionError:
                print(char_diff(actual_output_data, expected_output))

def assert_output_matches(filter_alias, doc_contents, expected_regex, ext=".txt"):
    if not ext.startswith("."):
        raise Exception("ext arg to assert_in_output must start with dot")

    with runfilter(filter_alias, doc_contents, ext=ext) as doc:
        if expected_regex:
            assert re.match(expected_regex, str(doc.output_data()))
        else:
            raise Exception(str(doc.output_data()))

def assert_output_cached(filter_alias, doc_contents, ext=".txt", min_filesize=None):
    if not ext.startswith("."):
        raise Exception("ext arg to assert_output_cached must start with dot")

    with runfilter(filter_alias, doc_contents, ext=ext) as doc:
        if doc.wrapper.state == 'ran':
            assert doc.output_data().is_cached()
            if min_filesize:
                assert doc.output_data().filesize() > min_filesize
        elif doc.wrapper.state == 'error':
            if isinstance(doc.wrapper.error, InactivePlugin):
                raise SkipTest()
            else:
                raise doc.wrapper.error
        else:
            raise Exception("state is '%s'" % doc.wrapper.state)

def assert_in_output(filter_alias, doc_contents, expected_output, ext=".txt"):
    if not ext.startswith("."):
        raise Exception("ext arg to assert_in_output must start with dot")

    with runfilter(filter_alias, doc_contents, ext=ext) as doc:
        if expected_output:
            actual_output = str(doc.output_data())
            msg = "did not find expected '%s' in actual output '%s'"
            assert expected_output in actual_output, msg % (expected_output, actual_output)
        else:
            raise Exception(str(doc.output_data()))

class capture_stdout():
    def __enter__(self):
        self.old_stdout = sys.stdout
        self.my_stdout = StringIO()
        sys.stdout = self.my_stdout
        return self.my_stdout

    def __exit__(self, type, value, traceback):
        sys.stdout = self.old_stdout
        self.my_stdout.close()

class capture_stderr():
    def __enter__(self):
        self.old_stderr = sys.stderr
        self.my_stderr = StringIO()
        sys.stderr = self.my_stderr
        return self.my_stderr

    def __exit__(self, type, value, traceback):
        sys.stderr = self.old_stderr
        self.my_stderr.close()

class run_templating_plugin():
    def __init__(self, klass, mock_attrs=None):
        if not mock_attrs:
            mock_attrs = {}
        self.f = MagicMock(**mock_attrs)
        self.plugin = klass(self.f)

    def __enter__(self):
        env = self.plugin.run()
        assert isinstance(env, dict)
        return env

    def __exit__(self, type, value, traceback):
        pass



from dexy.doc import Doc
from nose import SkipTest
from tests.utils import assert_output
from tests.utils import wrap

FORTRAN_HELLO_WORLD = """program hello
   print *, "Hello World!"
end program hello
"""

CPP_HELLO_WORLD = """#include <iostream>
using namespace std;

int main()
{
	cout << "Hello, world!";

	return 0;

}
"""

C_HELLO_WORLD = """#include <stdio.h>

int main()
{
    printf("HELLO, world\\n");
}
"""

C_FUSSY_HELLO_WORLD = """#include <stdio.h>

int main()
{
    printf("HELLO, world\\n");
    return 0;
}
"""

C_WITH_INPUT = """#include <stdio.h>

int main()
{
    int c;

    c = getchar();
    while (c != EOF) {
        putchar(c);
        c = getchar();
    }
}
"""
def test_fortran_filter():
    assert_output('fortran', FORTRAN_HELLO_WORLD, "Hello, world!", ext=".f")

def test_cpp_filter():
    assert_output('cpp', CPP_HELLO_WORLD, "Hello, world!", ext=".cpp")

def test_clang_filter():
    assert_output('clang', C_HELLO_WORLD, "HELLO, world\n", ext=".c")

def test_c_filter():
    assert_output('gcc', C_HELLO_WORLD, "HELLO, world\n", ext=".c")
    assert_output('gcc', C_FUSSY_HELLO_WORLD, "HELLO, world\n", ext=".c")

def test_cfussy_filter():
    raise SkipTest()
    assert_output('cfussy', C_FUSSY_HELLO_WORLD, "HELLO, world\n", ext=".c")
    with wrap() as wrapper:
        wrapper.debug = False
        doc = Doc("hello.c|cfussy",
                contents=C_HELLO_WORLD,
                wrapper=wrapper)
        wrapper.run_docs(doc)
        assert wrapper.state == 'error'

def test_c_input():
    with wrap() as wrapper:
        node = Doc("copy.c|cinput",
                inputs = [
                Doc("input.txt",
                    contents = "hello, c",
                    wrapper=wrapper)
                ],
                contents = C_WITH_INPUT,
                wrapper=wrapper)

        wrapper.run_docs(node)
        assert str(node.output_data()) == "hello, c"

def test_clang_input():
    with wrap() as wrapper:
        node = Doc("copy.c|clanginput",
                inputs = [
                Doc("input.txt",
                    contents = "hello, c",
                    wrapper=wrapper)
                ],
                contents = C_WITH_INPUT,
                wrapper=wrapper)

        wrapper.run_docs(node)
        assert str(node.output_data()) == "hello, c"

def test_clang_multiple_inputs():
    with wrap() as wrapper:
        node = Doc("copy.c|clanginput",
                inputs = [
                    Doc("input1.txt",
                        contents = "hello, c",
                        wrapper=wrapper),
                    Doc("input2.txt",
                        contents = "more data",
                        wrapper=wrapper)
                ],
                contents = C_WITH_INPUT,
                wrapper=wrapper)

        wrapper.run_docs(node)
        assert str(node.output_data()['input1.txt']) == 'hello, c'
        assert str(node.output_data()['input2.txt']) == 'more data'



from dexy.exceptions import UserFeedback
from dexy.filter import DexyFilter
import json
import mimetypes
import os
import requests

class ConfluenceRESTAPI(DexyFilter):
    """
    Filter for the Confluence REST API.
    """
    aliases = ['cfl', 'confluence']

    _settings = {
            'publish' : False,
            'input-extensions' : ['.html'],
            'output-extensions' : ['.json', '.html'],
            'url-base' : ("Root of URL.", None),
            'upload-attachments' : ("If True, attachments will automatically be uploaded. Can also be a list of file extensions.", True),
            'skip-attachments' : ("A list of file extensions which should not be uploaded as attachments.", None),
            'fix-attachment-paths' : ("If True, automatically replace attachement filenames with uploaded paths in document content.", True),
            'attachment-minor-edit' : ("Whether an attachment should be uploaded as a minor edit.", True),
            'attachment-comment' : ("Comment for attachment.", "Uploaded by Dexy confluence filter."),
            'wiki-path' : ("Path to wiki root.", "/wiki"),
            'api-path' : ("Path from wiki root to API endpoint.", "/rest/api"),
            'page-minor-edit' : ("Whether to mark page changes as a minor edit.", True),
            'space-key' : ("Confluence space in which to publish page.", None),
            'page-title' : ("Title of page to publish.", None),
            'page-id' : ("ID of existing Confluence page to update.", None),
            'parent-page-id': ("ID of page to use as parent.", None),
            'authstring' : ("base64-encoded username:password string.", "$CONFLUENCE_AUTHSTRING"),
            'username' : ("Confluence account username.", "$CONFLUENCE_USERNAME"),
            'password' : ("Confluence account password.", "$CONFLUENCE_PASSWORD"),
            'custom-mime-types' : ("Map of file extensions to mime types to supplement the python mimetypes module.", None)
            }

    def wiki_root_url(self):
        url_base = self.setting('url-base')
        if url_base is None:
            raise UserFeedback("The url-base setting must be provided.")
        if not url_base.startswith("http"):
            raise UserFeedback("The url-base setting should start with https.")
        return "%s%s" % (url_base, self.setting('wiki-path'))

    def url_for_path(self, path):
        if not path.startswith("/"):
            path = "/%s" % path
        return "%s%s%s" % (self.wiki_root_url(), self.setting('api-path'), path)

    def credentials(self):
        try:
            authstring = self.setting('authstring')
        except UserFeedback:
            authstring = None

        try:
            username = self.setting('username')
            password = self.setting('password')
        except UserFeedback:
            username = None
            password = None

        if authstring is not None:
            return { "headers" : { "Authorization" : "Basic %s" % authstring }}
        elif username is not None and password is not None:
            return { "auth" : (username, password) }
        else:
            raise UserFeedback("Must provide an authstring or both username and password.")

    def credentials_with_headers(self, headers):
        credentials = self.credentials()
        if 'headers' in credentials:
            credentials['headers'] = {}
        credentials['headers'].update(headers)
        return credentials

    def credentials_with_json_content_type(self):
        json_content_type = {'content-type': 'application/json'}
        return self.credentials_with_headers(json_content_type)

    def handle_response_code(self, response):
        if response.status_code < 400:
            pass
        elif response.status_code in range(400,500):
            raise UserFeedback(response.json()['message'])
        elif response.status_code in (500,):
            raise Exception("\nServer error %s:\n%s" % (response.status_code, response.json()['message']))
        else:
            print(response.text)
            raise Exception("Not set up to handle status code %s" % response.status_code)

    def get_path(self, path, params=None):
        response = requests.get(
                self.url_for_path(path),
                params=params,
                **self.credentials())
        self.handle_response_code(response)
        return response.json()

    def guess_mimetype(self, canonical_name):
        custom_mimetypes = self.setting('custom-mime-types')
        if custom_mimetypes is None:
            custom_mimetypes = {}
        ext = ".%s" % os.path.splitext(canonical_name)[1]
        if ext in custom_mimetypes:
            return custom_mimetypes[ext]
        else:
            mimetype, _ = mimetypes.guess_type(canonical_name)
            return mimetype

    def attachment_minor_edit_setting(self):
        return self.bool_setting_to_string(
                self.setting('attachment-minor-edit'))

    def page_minor_edit_setting(self):
        return self.bool_setting_to_string(
                self.setting('page-minor-edit'))

    def bool_setting_to_string(self, setting):
        if isinstance(setting, bool):
            if setting:
                return "true"
            else:
                return "false"
        else:
            return setting

    def post_file(self, path, canonical_name, filepath):
        no_check = {"X-Atlassian-Token" : "no-check"}
        mimetype = self.guess_mimetype(canonical_name)
        with open(filepath, 'rb') as fileref:
            files = {
                    'file': (canonical_name, fileref, mimetype),
                    'comment' : str(self.setting('attachment-comment')),
                    'minorEdit' : self.attachment_minor_edit_setting()
                    }
            response = requests.post(
                    self.url_for_path(path),
                    files=files,
                    **self.credentials_with_headers(no_check))
        self.handle_response_code(response)
        return response.json()

    def delete_path(self, path):
        response = requests.delete(
                self.url_for_path(path),
                **self.credentials())
        self.handle_response_code(response)
        return response.json()

    def json_post_path(self, path, data=None):
        response = requests.post(
                self.url_for_path(path),
                data=json.dumps(data),
                **self.credentials_with_json_content_type())
        self.handle_response_code(response)
        return response.json()

    def json_put_path(self, path, data=None):
        response = requests.put(
                self.url_for_path(path),
                data=json.dumps(data),
                **self.credentials_with_json_content_type())
        self.handle_response_code(response)
        return response.json()

    def find_page_info_by_title(self, page_title=None):
        space_key = self.setting('space-key')
        if page_title is None:
            page_title = self.setting('page-title')

        if space_key is None:
            raise UserFeedback("A space-key must be provided.")
        if page_title is None:
            raise UserFeedback("A page-title must be provided")

        params = {"spaceKey" : space_key, "title" : page_title }
        matching_pages = self.get_path("content", params)

        if matching_pages['size'] == 1:
            return matching_pages['results'][0]
        elif matching_pages['size'] == 0:
            return None
        else:
            print(matching_pages)
            raise Exception("Should only be 0 or 1 matching pages.")

    def find_page_id_by_title(self, page_title=None):
        page_info = self.find_page_info_by_title(page_title)
        if page_info is not None:
            print("Page found using title and space key, you should set page-id parameter to %s for greater robustness." % page_info['id'])
            return page_info['id']

    def ancestors(self):
        if self.setting('parent-page-id') is not None:
            return [{"type" : "page", "id" : self.setting('parent-page-id')}]
        else:
            return []

    def create_new_page(self):
        data = {
                "type" : "page",
                "title" : self.setting('page-title'),
                "ancestors" : self.ancestors(),
                "space" : {"key" : self.setting('space-key') },
                "body" : {
                    "storage" : {
                        "representation" : "storage",
                        "value" : str(self.input_data)
                    }
                  }
                }
        return self.json_post_path("content", data)

    def update_existing_page(self, page_id, page_text):
        page_info_args = {"expand" : "version"}
        page_info = self.get_path("content/%s" % page_id, page_info_args)
        page_version = page_info['version']['number']

        page_title = self.setting('page-title')
        if page_title is None:
            page_title = page_info['title']

        data = {
                "type" : "page",
                "title" : page_title,
                "ancestors" : self.ancestors(),
                "body" : {
                    "storage" : {
                        "representation" : "storage",
                        "value" : page_text
                    }
                  },
                'version' : {
                    'number' : page_version + 1,
                    'minorEdit' : self.bool_setting_to_string(
                        self.setting('page-minor-edit'))
                    }
                }
        return self.json_put_path("content/%s" % page_id, data)

    def upload_attachments(self, page_id):
        path = "content/%s/child/attachment" % page_id
        existing_attachments = self.get_path(path)['results']
        attachment_ids = dict(
                (att['title'], att['id'],)
                for att in existing_attachments)

        attachments = {}

        for input_doc in self.doc.walk_input_docs():
            if not input_doc.output_data().is_canonical_output():
                self.log_debug("Not uploading %s, set output to True if you want it." % input_doc)
                continue

            canonical_name = input_doc.output_data().basename()
            ext = os.path.splitext(canonical_name)[1]

            if not isinstance(self.setting('upload-attachments'), bool):
                if not ext in self.setting('upload-attachments'):
                    self.log_info("Skipping %s because %s not in %s" % (canonical_name, ext, self.setting('upload-attachments'),))
                    continue

            if self.setting('skip-attachments') is not None:
                if ext in self.setting('skip-attachments'):
                    self.log_info("Skipping %s because %s in %s" % (canonical_name, ext, self.setting('skip-attachments'),))
                    continue

            filepath = input_doc.output_data().storage.data_file()

            if canonical_name in attachment_ids:
                update_path = "%s/%s/data" % (path, attachment_ids[canonical_name])
                attachment = self.post_file(update_path, canonical_name, filepath)
                links = attachment['_links']
            else:
                attachment = self.post_file(path, canonical_name, filepath)
                links = attachment['results'][0]['_links']

            attachments[canonical_name] = links['download'].split("?")[0]

        return attachments

    def find_page_id(self):
        if self.setting('page-id') is not None:
            return self.setting('page-id')
        else:
            return self.find_page_id_by_title()

    def save_result(self, page):
        result = {}
        result['page-id'] = page['id']
        result['version'] = page['version']['number']
        result['url'] = "%s%s" % (self.wiki_root_url(), page['_links']['webui'])
        result['short-url'] = "%s%s" % (self.wiki_root_url(), page['_links']['tinyui'])
        try:
            self.output_data.set_data(json.dumps(result))
        except Exception as e:
            self.log_debug(e)
            self.log_debug(e.__class__.__name__)
            self.log_debug(result)

    def fix_attachment_paths(self, page_id, attachments):
        input_text = str(self.input_data)
        fixed_text = input_text

        for canonical_name, path in attachments.items():
            fullpath = self.wiki_root_url() + path
            fixed_text = fixed_text.replace(canonical_name, fullpath)

        return self.update_existing_page(page_id, fixed_text)

    def process(self):
        if not self.setting('publish'):
            self.output_data.set_data(str(self.input_data))
            print("not publishing", self.setting('space-key'), self.setting('page-title'))
            return
        else:
            print("uploading", self.setting('space-key'), self.setting('page-title'))

        page_id = self.find_page_id()

        if page_id is None:
            page = self.create_new_page()
            print("New page created. You should now set page-id parameter to %s." % page['id'])
        else:
            page = self.update_existing_page(page_id, str(self.input_data))

        if self.setting('upload-attachments'):
            attachments = self.upload_attachments(page['id'])
        else:
            attachments = {}

        if attachments and self.setting('fix-attachment-paths'):
            page = self.fix_attachment_paths(page['id'], attachments)

        self.save_result(page)



from dexy.wrapper import Wrapper
from tests.utils import TEST_DATA_DIR
from tests.utils import tempdir
from tests.utils import wrap
import dexy.doc
import dexy.node
import inspect
import os
import random
import shutil

LOGLEVEL = "WARN"

def assert_node_state(node, expected, additional_info=''):
    msg = "'%s' not in state '%s',  in state '%s'. %s"
    msgargs = (node.key, expected, node.state, additional_info)
    assert node.state == expected, msg % msgargs

def test_example_project():
    with tempdir():
        def run_from_cache_a_bunch_of_times():
            n = random.randint(2, 10)
            print("running %s times:" % n)
            for i in range(n):
                print('', i+1)
                wrapper = Wrapper(log_level=LOGLEVEL, debug=True)
                wrapper.run_from_new()
    
                for node in list(wrapper.nodes.values()):
                    assert_node_state(node, 'consolidated', "In iter %s" % i)

                wrapper.report()

        example_src = os.path.join(TEST_DATA_DIR, 'example')
        shutil.copytree(example_src, "example")
        os.chdir("example")

        wrapper = Wrapper(log_level=LOGLEVEL)
        wrapper.create_dexy_dirs()

        wrapper.run_from_new()
        wrapper.report()

        for node in list(wrapper.nodes.values()):
            assert_node_state(node, 'ran', "in first run")

        run_from_cache_a_bunch_of_times()

        # touch this file so it triggers cache updating
        os.utime("multiply.py", None)

        unaffected_keys = ('latex', 'pygments.sty|pyg', 's1/loop.py|pycon', 's1/loop.py|py',
                'main.rst|idio|h', 'main.rst|idio|l', 'main.rst|pyg|l', 'main.rst|pyg|h',
                's1/loop.py|idio|h', 's1/loop.py|idio|l', 's1/loop.py|pyg|l', 's1/loop.py|pyg|h',
                'dexy.yaml|idio|h', 'dexy.yaml|idio|l', 'dexy.yaml|pyg|l', 'dexy.yaml|pyg|h',
                )

        affected_keys = ('code', 'docs', "*|pyg|l", "*|pyg|h", "*|idio|l", "*|idio|h",
                "main.rst|jinja|rst|latex", "*.rst|jinja|rst|latex",
                "*.py|pycon", "*.py|py", "main.rst|jinja|rstbody|easyhtml",
                "*.rst|jinja|rstbody|easyhtml", "foo.txt",
                "multiply.py|idio|h", "multiply.py|idio|l", "multiply.py|pycon", "multiply.py|py",
                "multiply.py|pyg|h", "multiply.py|pyg|l",
                )

        wrapper = Wrapper(log_level=LOGLEVEL)
        wrapper.run_from_new()
        wrapper.report()

        for node in list(wrapper.nodes.values()):
            if node.key in unaffected_keys:
                assert_node_state(node, 'consolidated', "after touching multiply.py")
            else:
                assert node.key in affected_keys, node.key
                assert_node_state(node, 'ran', "after touchimg multiply.py")

        run_from_cache_a_bunch_of_times()

        import time
        time.sleep(0.5)

        with open("multiply.py", "r") as f:
            old_content = f.read()
        
        with open("multiply.py", "w") as f:
            f.write("raise")

        wrapper = Wrapper(log_level=LOGLEVEL)
        wrapper.run_from_new()
        assert wrapper.state == 'error'

        import time
        time.sleep(0.9)

        with open("multiply.py", "w") as f:
            f.write(old_content)

        wrapper = Wrapper(log_level=LOGLEVEL)
        wrapper.run_from_new()

        for node in list(wrapper.nodes.values()):
            if node.key in unaffected_keys:
                assert_node_state(node, 'consolidated', "after restoring old multiply.py content")
            else:
                assert node.key in affected_keys, node.key
                assert_node_state(node, 'ran', "after restoring old multiply.py contnet")

        wrapper.remove_dexy_dirs()
        wrapper.remove_reports_dirs(keep_empty_dir=True)
        wrapper.create_dexy_dirs()

        assert len(os.listdir(".dexy")) == 1

        wrapper = Wrapper(log_level=LOGLEVEL, dry_run=True)
        wrapper.run_from_new()
        wrapper.report()

        assert len(os.listdir(".dexy")) == 6

        with open(".dexy/reports/graph.txt", "r") as f:
            graph_text = f.read()

        assert "BundleNode(docs) (uncached)" in graph_text

        os.chdir("..")

def test_ragel_state_chart_to_image():
    ragel = inspect.cleandoc("""
        %%{
          machine hello_and_welcome;
          main := ( 'h' @ { puts "hello world!" }
                  | 'w' @ { puts "welcome" }
                  )*;
        }%%
          data = 'whwwwwhw'
          %% write data;
          %% write init;
          %% write exec;
        """)
    with wrap() as wrapper:
        graph_png = dexy.doc.Doc("example.rl|rlrbd|dot",
                wrapper,
                [],
                contents=ragel
                )

        syntax = dexy.doc.Doc("example.rl|rlrbd|pyg",
                wrapper,
                [],
                contents=ragel
                )

        wrapper.run_docs(graph_png, syntax)
        assert graph_png.state == 'ran'
        assert syntax.state == 'ran'



from dexy.data import Sectioned
from dexy.exceptions import UserFeedback
from tests.utils import wrap
import os

def test_create_new_sectioned_data():
    with wrap() as wrapper:
        settings = {
                'canonical-name' : "doc.txt"
                }
        data = Sectioned("doc.txt", ".txt", "def123", settings, wrapper)
        data.setup()

        data['alpha'] = "This is the first section."
        data['alpha']['abc'] = 123
        assert str(data['alpha']) == "This is the first section."

        data['beta'] = "This is the second section."
        data['beta']['def'] = 456

        assert str(data['alpha']) == "This is the first section."
        assert str(data['beta']) == "This is the second section."
        assert data['alpha']['abc'] == 123
        assert data['beta']['def'] == 456

        data['gamma'] = "This is the third section."
        del data['beta']
        assert list(data.keys()) == ['alpha', 'gamma']

def test_load_json():
    with wrap() as wrapper:
        os.makedirs(".dexy/this/de")
        with open(".dexy/this/de/def123.txt", "w") as f:
            f.write("""
            [
                { "foo" : "bar" },
                { "name" : "alpha", "contents" : "This is the first section.", "abc" : 123 } ,
                { "name" : "beta", "contents" : "This is the second section.", "def" : 456 } 
            ]
            """)

        settings = {
                'canonical-name' : "doc.txt"
                }
        data = Sectioned("doc.txt", ".txt", "def123", settings, wrapper)
        data.setup_storage()

        assert list(data.keys()) == ["alpha", "beta"]
        assert str(data) == "This is the first section.\nThis is the second section."
        assert str(data) == "This is the first section.\nThis is the second section."
        assert data.keyindex("alpha") == 0
        assert data.keyindex("beta") == 1
        assert data.keyindex("gamma") == -1

        assert str(data["alpha"]) == "This is the first section."
        assert str(data["beta"]) == "This is the second section."
        assert str(data["alpha"]) == "This is the first section."
        assert str(data["beta"]) == "This is the second section."

        assert data["foo"] == "bar"

        assert data['alpha']['abc'] == 123
        assert data['beta']['def'] == 456

        try:
            data["zxx"]
            assert False, "should raise error"
        except UserFeedback as e:
            assert "No value for zxx" in str(e)



from dexy.reporter import Reporter

class PlainTextGraph(Reporter):
    """
    Emits a plain text graph of the network structure.
    """
    aliases = ['graph']
    _settings = {
            'in-cache-dir' : True,
            'filename' : 'graph.txt',
            "run-for-wrapper-states" : ["ran", "checked"]
            }

    def run(self, wrapper):
        self.wrapper = wrapper

        def print_inputs(node, indent=0):
            content = []

            s = " " * indent * 4
            content.append("%s%s (%s)" % (s, node, node.state))

            for child in list(node.inputs) + node.children:
                content.extend(print_inputs(child, indent+1))
            return content

        graph = []
        for node in wrapper.roots:
            graph.extend(print_inputs(node))

        self.create_cache_reports_dir()
        with open(self.report_file(), "w") as f:
            f.write("\n".join(graph) + "\n")



from dexy.filters.process import SubprocessFilter
from dexy.utils import file_exists
import codecs
import dexy.exceptions
import dexy.utils
import os
import subprocess

class LatexFilter(SubprocessFilter):
    """
    Generates a PDF file from LaTeX source.
    """
    aliases = ['latex', 'pdflatex']
    _settings = {
            'executable' : 'pdflatex',
            'output' : True,
            'input-extensions' : ['.tex', '.txt'],
            'output-extensions' : ['.pdf'],
            'run-bibtex' : ("Should we run bibtex if a .bib file is an input?", True),
            'times-to-run-latex' : ("""How many times to run latex? (Latex is
                run one additional time if bibtex runs.)""", 2),
            'command-string' : "%(prog)s -interaction=nonstopmode -halt-on-error %(args)s %(script_file)s"
            }

    def process(self):
        self.populate_workspace()

        wd = self.parent_work_dir()
        env = self.setup_env()

        latex_command = self.command_string()

        if any(doc.output_data().ext == '.bib' for doc in self.doc.walk_input_docs()):
            bibtex_command = "bibtex %s" % os.path.splitext(self.output_data.basename())[0]
        else:
            bibtex_command = None

        def run_cmd(command):
            self.log_debug("running %s in %s" % (command, os.path.abspath(wd)))
            proc = subprocess.Popen(command, shell=True,
                                    cwd=wd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)

            stdout, stderr = proc.communicate()
            self.log_debug(stdout)

        if bibtex_command and self.setting('run-bibtex'):
            run_cmd(latex_command) #generate aux
            run_cmd(bibtex_command) #generate bbl

        n = self.setting('times-to-run-latex')
        for i in range(n):
            self.log_debug("running latex time %s (out of %s)" % (i+1, n))
            run_cmd(latex_command)

        if not file_exists(os.path.join(wd, self.output_data.basename())):
            log_file_path = os.path.join(wd, self.output_data.basename().replace(".pdf", ".log"))

            if file_exists(log_file_path):
                msg = "Latex file not generated. Look for information in latex log %s"
                msgargs = log_file_path
            else:
                msg = "Latex file not generated. Look for information in latex log in %s directory."
                msgargs = os.path.abspath(wd)

            raise dexy.exceptions.UserFeedback(msg % msgargs)

        if self.setting('add-new-files'):
            self.log_debug("adding new files found in %s for %s" % (wd, self.key))
            self.add_new_files()

        self.copy_canonical_file()

class TikzPdfFilter(LatexFilter):
    """
    Renders Tikz code to PDF.
    """
    aliases = ['tikz']

    def process(self):
        latex_filename = self.output_data.basename().replace(self.ext, ".tex")

        # TODO allow setting tikz libraries per-document, or just include all of them?
        # TODO how to create a page size that just includes the content
        latex_header = r"""\documentclass[tikz]{standalone}
\usetikzlibrary{shapes.multipart}
\begin{document}
        """
        latex_footer = "\n\end{document}"

        self.populate_workspace()
        wd = self.parent_work_dir()

        work_path = os.path.join(wd, latex_filename)
        self.log_debug("writing latex header + tikz content to %s" % work_path)
        with codecs.open(work_path, "w", encoding="utf-8") as f:
            f.write(latex_header)
            f.write(str(self.input_data))
            f.write(latex_footer)

        latex_command = "%s -interaction=batchmode %s" % (self.setting('executable'), latex_filename)

        def run_cmd(command):
            self.log_debug("about to run %s in %s" % (command, os.path.abspath(wd)))
            proc = subprocess.Popen(command, shell=True,
                                    cwd=wd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=self.setup_env())

            stdout, stderr = proc.communicate()

            if proc.returncode > 2: # Set at 2 for now as this is highest I've hit, better to detect whether PDF has been generated?
                raise dexy.exceptions.UserFeedback("latex error, look for information in %s" %
                                latex_filename.replace(".tex", ".log"))
            elif proc.returncode > 0:
                self.log_warn("""A non-critical latex error has occurred running %s,
                status code returned was %s, look for information in %s""" % (
                self.key, proc.returncode,
                latex_filename.replace(".tex", ".log")))

        run_cmd(latex_command)

        self.copy_canonical_file()



from dexy.filter import DexyFilter
from docutils import core
from docutils.frontend import OptionParser
from docutils.parsers.rst import Parser
from docutils.transforms import Transformer, frontmatter
from docutils.utils import new_document
import io
import dexy.exceptions
import docutils.writers
import os

def default_template(writer_name):
    """
    Set the default template correctly, in case there has been a change in working dir.
    """
    writer_class = docutils.writers.get_writer_class(writer_name)

    if os.path.isdir(writer_class.default_template_path):
        return os.path.abspath(os.path.join(writer_class.default_template_path, writer_class.default_template))
    else:
        return os.path.abspath(writer_class.default_template_path)

class RestructuredTextBase(DexyFilter):
    """ Base class for ReST filters using the docutils library.
    """
    aliases = []

    _settings = {
            "input-extensions" : [".rst", ".txt"],
            'output-extensions' : [".html", ".tex", ".xml", ".odt"],
            'output' : True,
            'writer' : ("Specify rst writer to use (not required: dexy will attempt to determine automatically from filename if not specified).", None),
            'stylesheet' : ("Stylesheet arg to pass to rst", None),
            'template' : ("Template arg to pass to rst", None),
            }

    def docutils_writer_name(self):
        if self.setting('writer'):
            return self.setting('writer')
        elif self.ext == ".html":
            return 'html'
        elif self.ext == ".tex":
            return 'latex2e'
        elif self.ext == ".xml":
            return 'docutils_xml'
        elif self.ext == ".odt":
            return 'odf_odt'
        else:
            raise Exception("unsupported extension %s" % self.ext)

class RestructuredText(RestructuredTextBase):
    """
    A 'native' ReST filter which uses the docutils library.

    Look for configuration options for writers here:
    http://docutils.sourceforge.net/docs/user/config.html
    """
    aliases = ['rst']
    skip_settings = 'settings-not-for-settings-overrides'
    _settings = {
            'allow-any-template-extension' : ("Whether to NOT raise an error if template extension does not match document extension.", False),
            skip_settings : (
                "Which of the settings should NOT be passed to settings_overrides.",
                ['writer']
                )
            }

    def process(self):
        def skip_setting(key):
            in_base_filter = key in DexyFilter._settings
            in_skip = key in self.setting(self.skip_settings) or key == self.skip_settings
            return in_base_filter or in_skip

        settings_overrides = dict((k.replace("-", "_"), v) for k, v in self.setting_values().items() if v and not skip_setting(k))
        writer_name = self.docutils_writer_name()

        warning_stream = io.StringIO()
        settings_overrides['warning_stream'] = warning_stream

        self.log_debug("settings for rst: %r" % settings_overrides)
        self.log_debug("rst writer: %s" % writer_name)

        # Check that template extension matches output.
        if 'template' in settings_overrides and not self.setting('allow-any-template-extension'):
            template = settings_overrides['template']
            template_ext = os.path.splitext(template)[1]
            if not template_ext == self.ext:
                msg = "You requested template '%s' with extension '%s' for %s, does not match document extension of '%s'"
                args = (template, template_ext, self.key, self.ext)
                raise dexy.exceptions.UserFeedback(msg % args)

        if not 'template' in settings_overrides:
            if hasattr(writer_name, 'default_template'):
                settings_overrides['template'] = default_template(writer_name)

        try:
            core.publish_file(
                    source_path = self.input_data.storage.data_file(),
                    destination_path = self.output_data.storage.data_file(),
                    writer_name=writer_name,
                    settings_overrides=settings_overrides
                    )
        except ValueError as e:
            if "Invalid placeholder in string" in e.message and 'template' in settings_overrides:
                self.log_warn("you are using template '%s'. is this correct?" % settings_overrides['template'])
            raise
        except Exception:
            self.log_warn("An error occurred while generating reStructuredText.")
            self.log_warn("source file %s" % (self.input_data.storage.data_file()))
            self.log_warn("settings for rst: %r" % settings_overrides)
            self.log_warn("rst writer: %s" % writer_name)
            raise

        self.log_debug("docutils warnings:\n%s\n" % warning_stream.getvalue())

class RstBody(RestructuredTextBase):
    """
    Returns just the body part of an ReST document.
    """
    aliases = ['rstbody']
    _settings = {
            'set-title' : ("Whether to set document title.", True),
            'output-extensions' : ['.html', '.tex']
            }

    def process_text(self, input_text):
        warning_stream = io.StringIO()
        settings_overrides = {}
        settings_overrides['warning_stream'] = warning_stream

        writer_name = self.docutils_writer_name()
        self.log_debug("about to call publish_parts with writer '%s'" % writer_name)

        if not 'template' in settings_overrides:
            settings_overrides['template'] = default_template(writer_name)

        try:
            parts = core.publish_parts(
                input_text,
                writer_name=writer_name,
                settings_overrides=settings_overrides
                )
        except AttributeError as e:
            raise dexy.exceptions.InternalDexyProblem(str(e))

        if self.setting('set-title') and ('title' in parts) and parts['title']:
            self.update_all_args({'title' : parts['title']})

        self.log_debug("docutils warnings:\n%s\n" % warning_stream.getvalue())

        return parts['body']

class RstMeta(RestructuredTextBase):
    """
    Extracts bibliographical metadata and makes this available to dexy.
    """
    aliases = ['rstmeta']
    _settings = {
            'output-extensions' : [".rst"]
            }

    def process_text(self, input_text):
        warning_stream = io.StringIO()
        settings_overrides = {}
        settings_overrides['warning_stream'] = warning_stream

        # Parse the input text using default settings
        settings = OptionParser(components=(Parser,)).get_default_values()
        parser = Parser()
        document = new_document('rstinfo', settings)
        parser.parse(input_text, document)

        # Transform the parse tree so that the bibliographic data is
        # is promoted from a mere field list to a `docinfo` node
        t = Transformer(document)
        t.add_transforms([frontmatter.DocTitle, frontmatter.DocInfo])
        t.apply_transforms()

        info = {}

        # Process individual nodes which are not part of docinfo.
        single_nodes = [
                docutils.nodes.title,
                docutils.nodes.subtitle,
                ]
        for node in single_nodes:
            for doc in document.traverse(node):
                if not len(doc.children) == 1:
                    msg = "Expected node %s to only have 1 child."
                    raise dexy.exceptions.InternalDexyProblem(msg % node)
                info[doc.tagname] = doc.children[0].astext()

        # Find the `docinfo` node and extract its children. Non-standard
        # bibliographic fields will have the `tagname` 'field' and two
        # children, the name and the value.  Standard fields simply keep
        # the name as the `tagname`.
        for doc in document.traverse(docutils.nodes.docinfo):
            for element in doc.children:
                if element.tagname == 'field':
                    name, value = element.children
                    name, value = name.astext(), value.astext()
                else:
                    name, value = element.tagname, element.astext()
                info[name] = value

        self.log_debug("found info:\n%s\n" % info)
        self.update_all_args(info)
        self.log_debug("docutils warnings:\n%s\n" % warning_stream.getvalue())

        return input_text

class RstDocParts(DexyFilter):
    """
    Returns key-value storage of document parts.
    """
    aliases = ['rstdocparts']
    _settings = {
            'input-extensions' : [".rst", ".txt"],
            'data-type' : 'keyvalue',
            'output-extensions' : ['.sqlite3', '.json'],
            'writer' : ("Specify rst writer to use.", 'html')
            }

    def process(self):
        input_text = str(self.input_data)

        warning_stream = io.StringIO()
        settings_overrides = {}
        settings_overrides['warning_stream'] = warning_stream

        writer_name = self.setting('writer')

        if not 'template' in settings_overrides:
            settings_overrides['template'] = default_template(writer_name)

        parts = core.publish_parts(
                input_text,
                writer_name=writer_name,
                settings_overrides=settings_overrides
                )

        self.log_debug("docutils warnings:\n%s\n" % warning_stream.getvalue())

        for k, v in parts.items():
            self.output_data.append(k, v)
        self.output_data.save()



from dexy.batch import Batch
from dexy.commands.utils import init_wrapper
from dexy.data import Generic
from dexy.data import KeyValue
from dexy.data import Sectioned
from dexy.utils import defaults
from operator import attrgetter
import dexy.exceptions
import json
import sys

def grep_command(
        __cli_options=False, # nodoc
        contents=False, # print out the contents of each matched file
        expr="", # An expression partially matching document name.
        key="", # An exact document key
        keyexpr="", # Only search for keys matching this expression
        keylimit=10, # Maximum number of matching keys to print
        keys=False, # List keys in documents
        limit=10, # Maximum number of matching documents to print
        lines=False, # maximum number of lines of content to print
        **kwargs
        ):
    """
    Search for documents and sections within documents.

    Dexy must have already run successfully.

    You can search for documents based on exact key or inexpect expression. The
    number of documents returned is controlled by --limit.

    You can print all keys in found documents by requesting --keys, number of
    results is controlled by --keylimit.

    You can search the section names/keys in found documents by passing a
    --keyexpr

    You can print contents of documents by requesting --contents, number of
    lines of content can be controlled by --lines.

    This does not search contents of documents, just document names and
    internal section names.
    """

    artifactsdir = kwargs.get('artifactsdir', defaults['artifacts_dir'])
    wrapper = init_wrapper(locals())
    batch = Batch.load_most_recent(wrapper)
   
    if not batch:
        print("you need to run dexy first")
        sys.exit(1)
    else:
        if expr:
            matches = sorted([data for data in batch if expr in data.key],
                    key=attrgetter('key'))
        elif key:
            matches = sorted([data for data in batch if key == data.key],
                    key=attrgetter('key'))
        else:
            raise dexy.exceptions.UserFeedback("Must specify either expr or key")

        n = len(matches)
        if n > limit:
            print("only printing first %s of %s total matches" % (limit, n))
            matches = matches[0:limit]

        for match in matches:
            print_match(match, keys, keyexpr, contents, keylimit, lines)

def print_match(match, keys, keyexpr, contents, keylimit, lines):
    print(match.key, "\tcache key:", match.storage_key)

    if hasattr(match, 'keys'):
        if keyexpr:
            print_keys([key for key in list(match.keys()) if keyexpr in key], keylimit, lines)
        elif keys:
            print_keys(list(match.keys()), keylimit, lines)

    if contents:
        if isinstance(match, Sectioned):
            for section_name, section_contents in match.data().items():
                print("  section: %s" % section_name)
                print()
                print_contents(section_contents, lines)
                print()
        elif isinstance(match, KeyValue):
            pass
        elif isinstance(match, Generic):
            try:
                json.dumps(str(match))
                print_contents(str(match), lines)
            except UnicodeDecodeError:
                print("  not printable")

def print_keys(pkeys, keylimit, lines):
    n = len(pkeys)
    if n > keylimit:
        pkeys = pkeys[0:keylimit]
    
    for key in pkeys:
        print('  ', key)

    if n > keylimit:
        print("  only printed first %s of %s total keys" % (keylimit, n))

def print_contents(text, lines):
    text_lines = text.splitlines()
    for i, line in enumerate(text_lines):
        if lines and i > lines-1:
            continue
        print("  ", line)

    if lines and lines < len(text_lines):
        print("   only printed first %s of %s total lines" % (lines, len(text_lines)))



from dexy.version import DEXY_VERSION
from dexy.wrapper import Wrapper
from io import StringIO
from mock import patch
from nose.exc import SkipTest
from nose.tools import raises
from tests.utils import tempdir
from tests.utils import wrap
import dexy.commands
import os
import sys

def test_init_wrapper():
    with tempdir():
        with open("dexy.conf", "w") as f:
            f.write("artifactsdir: custom")

        modargs = {}
        wrapper = dexy.commands.utils.init_wrapper(modargs)
        assert wrapper.artifacts_dir == 'custom'

@patch.object(sys, 'argv', ['dexy', 'setup'])
def test_setup_with_dexy_conf_file():
    with tempdir():
        with open("dexy.conf", "w") as f:
            f.write("artifactsdir: custom")

        dexy.commands.run()
        assert os.path.exists("custom")
        assert os.path.isdir("custom")
        assert not os.path.exists("artifacts")

@raises(SystemExit)
@patch.object(sys, 'argv', ['dexy', 'grep', '-expr', 'hello'])
def test_grep():
    with wrap():
        dexy.commands.run()

@patch.object(sys, 'argv', ['dexy', 'grep'])
@patch('sys.stderr', new_callable=StringIO)
def test_grep_without_expr(stderr):
    raise SkipTest()
    try:
        dexy.commands.run()
    except SystemExit as e:
        assert e.code == 1
        assert 'Must specify either expr or key' in stderr.getvalue()

@patch.object(sys, 'argv', ['dexy'])
@patch('sys.stderr', new_callable=StringIO)
def test_run_with_userfeedback_exception(stderr):
    with wrap():
        with open("docs.txt", "w") as f:
            f.write("*.py|py")

        with open("hello.py", "w") as f:
            f.write("raise")

        dexy.commands.run()

@patch.object(sys, 'argv', ['dexy', 'invalid'])
@patch('sys.stdout', new_callable=StringIO)
def test_run_invalid_command(stdout):
    try:
        dexy.commands.run()
        assert False, 'should raise SystemExit'
    except SystemExit as e:
        assert e.code == 1

@patch.object(sys, 'argv', ['dexy', '--help'])
@patch('sys.stdout', new_callable=StringIO)
def test_run_help_old_syntax(stdout):
    dexy.commands.run()
    assert "Commands for running dexy:" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', '--version'])
@patch('sys.stdout', new_callable=StringIO)
def test_run_version_old_syntax(stdout):
    dexy.commands.run()
    assert DEXY_VERSION in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'help'])
@patch('sys.stdout', new_callable=StringIO)
def test_run_help(stdout):
    dexy.commands.run()
    assert "Commands for running dexy:" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'version'])
@patch('sys.stdout', new_callable=StringIO)
def test_run_version(stdout):
    dexy.commands.run()
    assert DEXY_VERSION in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy'])
@patch('sys.stdout', new_callable=StringIO)
def test_run_dexy(stdout):
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        dexy.commands.run()

### "viewer-ping"
@patch.object(sys, 'argv', ['dexy', 'viewer:ping'])
@patch('sys.stdout', new_callable=StringIO)
def test_viewer_command(stdout):
    raise SkipTest()
    dexy.commands.run()
    assert "pong" in stdout.getvalue()

### "conf"
@patch.object(sys, 'argv', ['dexy', 'conf'])
@patch('sys.stdout', new_callable=StringIO)
def test_conf_command(stdout):
    with tempdir():
        dexy.commands.run()
        assert os.path.exists("dexy.conf")
        assert "has been written" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'conf'])
@patch('sys.stdout', new_callable=StringIO)
def test_conf_command_if_path_exists(stdout):
    with tempdir():
        with open("dexy.conf", "w") as f:
            f.write("foo")
        assert os.path.exists("dexy.conf")
        dexy.commands.run()
        assert "dexy.conf already exists" in stdout.getvalue()
        assert "artifactsdir" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'conf', '-p'])
@patch('sys.stdout', new_callable=StringIO)
def test_conf_command_with_print_option(stdout):
    with tempdir():
        dexy.commands.run()
        assert not os.path.exists("dexy.conf")
        assert "artifactsdir" in stdout.getvalue()

### "filters"
@patch.object(sys, 'argv', ['dexy', 'filters'])
@patch('sys.stdout', new_callable=StringIO)
def test_filters_cmd(stdout):
    dexy.commands.run()
    assert "pyg : Apply Pygments" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'filters', '-alias', 'pyg'])
@patch('sys.stdout', new_callable=StringIO)
def test_filters_cmd_alias(stdout):
    dexy.commands.run()
    assert "pyg, pygments" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'filters', '-versions'])
@patch('sys.stdout', new_callable=StringIO)
def test_filters_text_versions__slow(stdout):
    dexy.commands.run()
    assert "Installed version: Python" in stdout.getvalue()

@patch.object(sys, 'argv', ['dexy', 'filters', '-alias', 'pyg', '-source'])
@patch('sys.stdout', new_callable=StringIO)
def test_filters_text_single_alias_source(stdout):
    raise SkipTest() # TODO fixme
    dexy.commands.run()
    text = stdout.getvalue()
    sys.stderr.write("TEXT IS:\n%s" % text)
    assert "pyg, pygments" in text
    assert "class" in text
    assert "PygmentsFilter" in text
    assert not "class PygmentsFilter" in text

@patch.object(sys, 'argv', ['dexy', 'filters', '-alias', 'pyg', '-source', '-nocolor'])
@patch('sys.stdout', new_callable=StringIO)
def test_filters_text_single_alias_source_nocolor(stdout):
    raise SkipTest() # TODO fixme
    dexy.commands.run()
    text = stdout.getvalue()
    assert "pyg, pygments" in text
    assert "class PygmentsFilter" in text

@patch.object(sys, 'argv', ['dexy', 'parsers'])
@patch('sys.stdout', new_callable=StringIO)
def test_parsers_text(stdout):
    dexy.commands.run()
    text = stdout.getvalue()
    assert "Yaml Parser" in text

@patch.object(sys, 'argv', ['dexy', 'nodes'])
@patch('sys.stdout', new_callable=StringIO)
def test_nodes_text(stdout):
    dexy.commands.run()
    text = stdout.getvalue()
    assert "bundle" in text

@patch.object(sys, 'argv', ['dexy', 'env'])
@patch('sys.stdout', new_callable=StringIO)
def test_env_text(stdout):
    dexy.commands.run()
    text = stdout.getvalue()
    assert "uuid" in text



import os
from dexy.doc import Doc
from tests.utils import wrap

def test_output_reporter():
    with wrap() as wrapper:
        wrapper.reports = "output"
        doc = Doc("hello.txt", wrapper, [], contents="hello")
        wrapper.run_docs(doc)
        wrapper.report()
        assert os.path.exists("output")
        assert os.path.exists("output/hello.txt")



import cashew
from cashew import Plugin

class PluginMeta(cashew.PluginMeta):
    """
    PluginMeta customized for dexy.
    """
    _store_other_class_settings = {} # allow plugins to define settings for other classes
    official_dexy_plugins = ("dexy_templates", "dexy_viewer", "dexy_filter_examples")

    def load_class_from_locals(cls, class_name):
        from dexy.template import Template
        from dexy.filters.pexp import PexpectReplFilter
        from dexy.filters.process import SubprocessCompileFilter
        from dexy.filters.process import SubprocessCompileInputFilter
        from dexy.filters.process import SubprocessExtToFormatFilter
        from dexy.filters.process import SubprocessFilter
        from dexy.filters.process import SubprocessFormatFlagFilter
        from dexy.filters.process import SubprocessInputFileFilter
        from dexy.filters.process import SubprocessInputFilter
        from dexy.filters.process import SubprocessStdoutFilter
        from dexy.filters.process import SubprocessStdoutTextFilter
        from dexy.filters.standard import PreserveDataClassFilter
        return locals()[class_name]

    def apply_prefix(cls, modname, alias):
        if modname.startswith("dexy_") and not modname in PluginMeta.official_dexy_plugins:
            prefix = modname.split(".")[0].replace("dexy_", "")
            return "%s:%s" % (prefix, alias)
        else:
            return alias 

    def adjust_alias(cls, alias):
        """
        All of '-' or '-foo' or '---' should map to '-' which is a registered alias.

        This way we can always create unique names by including arbitrary
        distinguishing content after a '-'.
        """
        if alias.startswith('-'):
            alias = '-'
        return alias

class Command(Plugin, metaclass=PluginMeta):
    """
    Parent class for custom dexy commands.
    """
    _settings = {}
    aliases = []
    default_cmd = None
    namespace = None

class TemplatePlugin(Plugin, metaclass=PluginMeta):
    """
    Parent class for template plugins.
    """
    aliases = []
    _settings = {
            'no-jinja-filter' : ("Listed entries should not be made into jinja filters.")
            }

    def is_active(self):
        return True

    def __init__(self, filter_instance=None):
        if filter_instance:
            self.filter_instance = filter_instance

    def run(self):
        return {}



from dexy.doc import Doc
from tests.utils import assert_output
from tests.utils import wrap
import json
import os

def test_header_footer_filters():
    with wrap() as wrapper:
        os.makedirs('subdir/subsubdir')
        node = Doc("subdir/file.txt|hd|ft",
                wrapper,
                [
                    Doc("_header.txt", wrapper, [], contents="This is a header in parent dir."),
                    Doc("subdir/_header.txt|jinja", wrapper, [], contents="This is a header."),
                    Doc("subdir/_footer.txt|jinja", wrapper, [], contents="This is a footer."),
                    Doc("subdir/subsubdir/_header.txt", wrapper, [], contents="This is a header in a subdirectory.")
                    ],
                contents="These are main contents."
                )

        wrapper.run_docs(node)
        assert str(node.output_data()) == "This is a header.\nThese are main contents.\nThis is a footer."

def test_join_filter():
    contents = json.loads("""[{},
    {"name" : "1", "contents" : "section one" },
    {"name" : "2", "contents" : "section two" }
    ]""")
    assert_output("join", contents, "section one\nsection two")

def test_head_filter():
    assert_output("head", "1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n", "1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n")

def test_word_wrap_filter():
    with wrap() as wrapper:
        node = Doc("example.txt|wrap", wrapper, [], contents="this is a line of text", wrap={"width" : 5})
        wrapper.run_docs(node)
        assert str(node.output_data()) == "this\nis a\nline\nof\ntext"

def test_lines_filter():
    expected = {}
    expected['1'] = "line one"
    expected['2'] = "line two"
    assert_output("lines", "line one\nline two", expected)

def test_ppjson_filter():
    assert_output(
            "ppjson",
            '{"foo" :123, "bar" :456}',
            """{\n    "bar": 456, \n    "foo": 123\n}"""
            )

def test_start_space_filter():
    o = {}
    o['1'] = " abc\n def"
    assert_output("startspace", "abc\ndef", o)

def test_tags_filter():
    with wrap() as wrapper:
        node = Doc("example.txt|tags", wrapper, [], contents="<p>the text</p>", tags={"tags" : ["html", "body"]})
        wrapper.run_docs(node)
        assert str(node.output_data()) == "<html><body>\n<p>the text</p>\n</body></html>"




from dexy.version import DEXY_VERSION
import datetime

citation_formats = ['bibtex']

def build_cite_parser(parser):
    parser.set_defaults(cmd=cite_command)
    parser.add_argument('-f', '--fmt', default=citation_formats[0], help="Desired format of citation.")

# TODO list available citation types

def cite_command(args):
    fmt = args.fmt
    if fmt == 'bibtex':
        cite_bibtex()
    else:
        msg = f"Don't know how to provide citation in '%{fmt}' format"
        print(msg)
        # raise dexy.exceptions.UserFeedback(msg % fmt)

def bibtex_text():
    args = {
            'version' : DEXY_VERSION,
            'year' : datetime.date.today().year
            }

    return """@misc{Dexy,
    title = {Dexy: Reproducible Data Analysis and Document Automation Software, Version~%(version)s},
    author = {{Nelson, Ana}},
    year = {%(year)s},
    url = {http://www.dexy.it/},
    note = {http://orcid.org/0000-0003-2561-1564}
}""" % args

def cite_bibtex():
    print((bibtex_text()))



from dexy.parser import Parser
from dexy.utils import parse_json

class Environment(Parser):
    """
    Parent class for environment parsers.
    """
    @classmethod
    def parse_environment_from_text(klass, text):
        pass

    def parse(self, parent_dir, config_text):
        config = self.parse_environment_from_text(config_text)
        self.ast.environment_for_directory.append((parent_dir, config,))

class JsonEnvironment(Environment):
    """
    Loads environment variables from a JSON file.
    """
    aliases = ['dexy-env.json']

    @classmethod
    def parse_environment_from_text(klass, text):
        return parse_json(text)

class PythonEnvironment(Environment):
    """
    Loads environment variables from a python script.
    """
    aliases = ['dexy-env.py']

    @classmethod
    def parse_environment_from_text(klass, text):
        env = {}
        skip = ('env', 'skip', 'self', 'parent_dir', 'env_text')

        exec(text)

        for k, v in locals().items():
            if not k in skip:
                env[k] = v
        return env



from tests.utils import wrap
from dexy.doc import Doc
from nose.exc import SkipTest

def test_pytest_filter():
    raise SkipTest() # this is running dexy's tests, not cashew's tests
    with wrap() as wrapper:
        doc = Doc(
                "modules.txt|pytest",
                wrapper,
                [],
                contents="cashew"
                )
        wrapper.run_docs(doc)
        data = doc.output_data()

        testname = "test_cashew.test_standardize_alias_or_aliases"
        assert data[testname + ':doc'] == "docstring for test"
        assert data[testname + ':name'] == "test_standardize_alias_or_aliases"
        assert data[testname + ':comments'] == "# comment before test\n"
        assert bool(data[testname + ':passed'])
        assert "def test_standardize_alias_or_aliases():" in data[testname + ':source']



import dexy.filters.ansi
import dexy.filters.api
import dexy.filters.archive
import dexy.filters.asciidoctor
import dexy.filters.aws
import dexy.filters.confluence
import dexy.filters.deprecated
import dexy.filters.easy
import dexy.filters.example
import dexy.filters.fluid_html
import dexy.filters.genipynb
import dexy.filters.git
import dexy.filters.id
import dexy.filters.ipynb
import dexy.filters.ipynbcasper
import dexy.filters.java
import dexy.filters.latex
import dexy.filters.lyx
import dexy.filters.md
import dexy.filters.org
import dexy.filters.pexp
import dexy.filters.phantomjs
import dexy.filters.pydoc
import dexy.filters.pytest
import dexy.filters.pyparse
import dexy.filters.pyg
import dexy.filters.pyn
import dexy.filters.rst
import dexy.filters.sanitize
import dexy.filters.soup
import dexy.filters.split
import dexy.filters.standard
import dexy.filters.sub
import dexy.filters.templating
import dexy.filters.websequence
import dexy.filters.wordpress
import dexy.filters.yamlargs
import dexy.filters.xxml

import dexy.filter
import os

yaml_file = os.path.join(os.path.dirname(__file__), 'filters.yaml')
dexy.filter.Filter.register_plugins_from_yaml_file(yaml_file)



from tests.utils import wrap
from dexy.doc import Doc

def test_pydoc_filter_on_module_names():
    with wrap() as wrapper:
        doc = Doc("modules.txt|pydoc", wrapper, [], contents="os math")
        wrapper.run_docs(doc)
        data = doc.output_data()
        assert len(list(data.keys())) > 100
        assert data['math.e:value'].startswith("2.71828")

python_file_content = """
import math

# Comment for foo
def foo():
    '''
    docstring for foo
    '''
    return True

# Comment for bar
def bar():
    '''
    docstring for bar
    '''
    return False

"""

def test_pydoc_filter_on_python_files():
    with wrap() as wrapper:
        doc = Doc("source.py|pydoc", wrapper, [], contents=python_file_content)
        wrapper.run_docs(doc)

        data = doc.output_data()
        keys = list(data.keys())

        assert 'bar:source' in keys
        assert 'foo:source' in keys
        
        assert data['foo:doc'] == "docstring for foo"
        assert data['foo:comments'] == "# Comment for foo\n"

        assert data['bar:doc'] == "docstring for bar"
        assert data['bar:comments'] == "# Comment for bar\n"
        



from dexy.commands.utils import init_wrapper
from dexy.utils import defaults
from operator import attrgetter
import dexy.exceptions
import os
import subprocess
import sys
import time

def dexy_command(
        __cli_options=False,
        artifactsdir=defaults['artifacts_dir'], # location of directory in which to store artifacts
        conf=defaults['config_file'], # name to use for configuration file
        configs=defaults['configs'], # list of doc config files to parse
        debug=defaults['debug'], # Prints stack traces, other debug stuff.
        directory=defaults['directory'], # Allow processing just a subdirectory.
        dryrun=defaults['dry_run'], # if True, just parse config and print batch info, don't run dexyT
        encoding=defaults['encoding'], # Default encoding. Set to 'chardet' to use chardet auto detection.
        exclude=defaults['exclude'], # comma-separated list of directory names to exclude from dexy processing
        excludealso=defaults['exclude_also'], # comma-separated list of directory names to exclude from dexy processing
        full=defaults['full'], # Whether to do a full run including tasks marked default: False
        globals=defaults['globals'], # global values to make available within dexy documents, should be KEY=VALUE pairs separated by spaces
        help=False, #nodoc
        h=False, #nodoc
        hashfunction=defaults['hashfunction'], # What hash function to use, set to crc32 or adler32 for more speed but less reliability
        include=defaults['include'], # Locations to include which would normally be excluded.
        logdir=defaults['log_dir'], # DEPRECATED
        logfile=defaults['log_file'], # name of log file
        logformat=defaults['log_format'], # format of log entries
        loglevel=defaults['log_level'], # log level, valid options are DEBUG, INFO, WARN
        nocache=defaults['dont_use_cache'], # whether to force dexy not to use files from the cache
        noreports=False, # if true, don't run any reports
        outputroot=defaults['output_root'], # Subdirectory to use as root for output
        pickle=defaults['pickle'], # library to use for persisting info to disk, may be 'c', 'py', 'json'
        plugins=defaults['plugins'], # additional python packages containing dexy plugins
        profile=defaults['profile'], # whether to run with cProfile. Arg can be a boolean, in which case profile saved to 'dexy.prof', or a filename to save to.
        r=False, # whether to clear cache before running dexy
        recurse=defaults['recurse'], # whether to include doc config files in subdirectories
        reports=defaults['reports'], # reports to be run after dexy runs, enclose in quotes and separate with spaces
        reset=False, # whether to clear cache before running dexy
        silent=defaults['silent'], # Whether to not print any output when running dexy
        strace=defaults['strace'], # Run dexy using strace (VERY slow)
        uselocals=defaults['uselocals'], # use cached local copies of remote URLs, faster but might not be up to date, 304 from server will override this setting
        target=defaults['target'], # Which target to run. By default all targets are run, this allows you to run only 1 bundle (and its dependencies).
        version=False, # For people who type -version out of habit
        writeanywhere=defaults['writeanywhere'] # Whether dexy can write files outside of the dexy project root.
    ):
    """
    Runs Dexy.
    """
    if h or help:
        return dexy.commands.help_command()

    if version:
        return dexy.commands.version_command()

    if r or reset:
        dexy.commands.dirs.reset_command(artifactsdir=artifactsdir, logdir=logdir)

    if silent:
        print("sorry, -silent option not implemented yet https://github.com/ananelson/dexy/issues/33")

    wrapper = init_wrapper(locals())
    wrapper.assert_dexy_dirs_exist()
    run_reports = (not noreports)

    try:
        if profile:
            run_dexy_in_profiler(wrapper, profile)

        elif strace:
            run_dexy_in_strace(wrapper, strace)
            run_reports = False

        else:
            start = time.time()
            wrapper.run_from_new()
            elapsed = time.time() - start
            print("dexy run finished in %0.3f%s" % (elapsed, wrapper.state_message()))

    except dexy.exceptions.UserFeedback as e:
        handle_user_feedback_exception(wrapper, e)

    except KeyboardInterrupt:
        handle_keyboard_interrupt()

    except Exception as e:
        log_and_print_exception(wrapper, e)
        raise

    if run_reports and hasattr(wrapper, 'batch'):
        start_time = time.time()
        wrapper.report()
        print("dexy reports finished in %0.3f" % (time.time() - start_time))

it_command = dexy_command

def log_and_print_exception(wrapper, e):
    if hasattr(wrapper, 'log'):
        wrapper.log.error("An error has occurred.")
        wrapper.log.error(e)
        wrapper.log.error(e.message)
    import traceback
    traceback.print_exc()

def handle_user_feedback_exception(wrapper, e):
    if hasattr(wrapper, 'log'):
        wrapper.log.error("A problem has occurred with one of your documents:")
        wrapper.log.error(e.message)
    sys.stderr.write("ERROR: Oops, there's a problem processing one of your documents. Here is the error message:" + os.linesep)
    for line in e.message.splitlines():
        sys.stderr.write("  " + line + "\n")
    if not e.message.endswith(os.linesep) or e.message.endswith("\n"):
        sys.stderr.write(os.linesep)

def handle_keyboard_interrupt():
    sys.stderr.write("""
    ok, stopping your dexy run
    you might want to 'dexy reset' before running again\n""")
    sys.exit(1)

def run_dexy_in_profiler(wrapper, profile):
    # profile may be a boolean or the name of a file to use
    if isinstance(profile, bool):
        profile_filename = os.path.join(wrapper.artifacts_dir, "dexy.prof")
    else:
        profile_filename = profile

    # run dexy in profiler
    import cProfile
    print("running dexy with cProfile, writing profile data to %s" % profile_filename)
    cProfile.runctx("wrapper.run_from_new()", None, locals(), profile_filename)

    # print report
    import pstats
    stats_output_file = os.path.join(wrapper.artifacts_dir, "profile-report.txt")

    with open(stats_output_file, 'w') as f:
        stat = pstats.Stats(profile_filename, stream=f)
        stat.sort_stats("cumulative")
        stat.print_stats()

    print("Report is in %s, profile data is in %s." % (stats_output_file, profile_filename))

def run_dexy_in_strace(wrapper, strace):
    if isinstance(strace, bool):
        strace_filename = 'dexy.strace'
    else:
        strace_filename = strace

    def run_command(command):
        proc = subprocess.Popen(
                   command,
                   shell=True,
                   stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE
                   )
        stdout, stderr = proc.communicate()
        print(stdout)

    commands = ( 
            "strace dexy --reports \"\" 2> %s" % strace_filename, # TODO pass command line args except for --strace option
            "echo \"calls to stat:\" ; grep \"^stat(\" %s | wc -l" % strace_filename,
            "echo \"calls to read:\" ; grep \"^read(\" %s | wc -l" % strace_filename,
            "echo \"calls to write:\" ; grep \"^write(\" %s | wc -l" % strace_filename,
            "grep \"^stat(\" %s | sort | uniq -c | sort -r -n > strace-stats.txt" % strace_filename,
            "grep \"^read(\" %s | sort | uniq -c | sort -r -n > strace-reads.txt" % strace_filename,
            "grep \"^write(\" %s | sort | uniq -c | sort -r -n > strace-writes.txt" % strace_filename,
        )

    for command in commands:
        run_command(command)

def targets_command(
        full=False, # Whether to just print likely pretty target names, or all names.
        **kwargs):
    """
    Prints a list of available targets, which can be run via "dexy -target name".
    """
    wrapper = init_wrapper(locals())
    wrapper.assert_dexy_dirs_exist()
    wrapper.to_valid()
    wrapper.to_walked()

    print("Targets you can pass to -target option:")
    for doc in sorted(wrapper.bundle_docs(), key=attrgetter('key')):
        print("  ", doc.key)

    if full:
        print()
        print("These targets are also available, with lower priority:")
        for doc in sorted(wrapper.non_bundle_docs(), key=attrgetter('key')):
            print("  ", doc.key)
        print()
        print("""Target names can be matched exactly or with the first few characters,
in which case all matching targets will be run.""")
    else:
        print()
        print("Run this command with --full option for additional available target names.")



from dexy.filter import DexyFilter
from dexy.plugin import TemplatePlugin
from jinja2 import FileSystemLoader
from jinja2.exceptions import TemplateNotFound
from jinja2.exceptions import TemplateSyntaxError
from jinja2.exceptions import UndefinedError
import dexy.exceptions
import jinja2
import jinja2.ext
import os
import re
import traceback

class PassThroughWhitelistUndefined(jinja2.StrictUndefined):
    call_whitelist = ('link', 'section',)

    def wrap_arg(self, arg):
        if isinstance(arg, str):
            return "\"%s\"" % str(arg)
        else:
            return str(arg)

    def __call__(self, *args, **kwargs):
        name = self._undefined_name

        if name in self.call_whitelist:
            msgargs = {
                    'name' : name,
                    'argstring' : ",".join(self.wrap_arg(a) for a in args)
                    }
            return "{{ %(name)s(%(argstring)s) }}" % msgargs
        else:
            self._fail_with_undefined_error(*args, **kwargs)

class TemplateFilter(DexyFilter):
    """
    Base class for templating system filters such as JinjaFilter. Templating
    systems are used to make generated artifacts available within documents.

    Plugins are used to prepare content.
    """
    aliases = ['template']

    _settings = {
            'output' : True,
            'variables' : ("Variables to be made available to document.", {}),
            'vars' : ("Variables to be made available to document.", {}),
            'plugins' : ("List of plugins for run_plugins to use.", []),
            'skip-plugins' : ("List of plugins which run_plugins should not use.", [])
            }

    def template_plugins(self):
        """
        Returns a list of plugin classes for run_plugins to use.
        """
        if self.setting('plugins'):
            return [TemplatePlugin.create_instance(alias, self)
                        for alias in self.setting('plugins')]
        else:
            return [instance for instance in TemplatePlugin.__iter__(self)
                        if not instance.alias in self.setting('skip-plugins')]

    def run_plugins(self):
        env = {}
        for plugin in self.template_plugins():
            self.log_debug("Running template plugin %s" % plugin.__class__.__name__)
            new_env_vars = plugin.run()
            if new_env_vars is None:
                msg = "%s did not return any values"
                raise dexy.exceptions.InternalDexyProblem(msg % plugin.alias)
            if any(v in list(env.keys()) for v in new_env_vars):
                new_keys = ", ".join(sorted(new_env_vars))
                existing_keys = ", ".join(sorted(env))
                msg = "plugin class '%s' is trying to add new keys '%s', already have '%s'"
                raise dexy.exceptions.InternalDexyProblem(msg % (plugin.__class__.__name__, new_keys, existing_keys))
            env.update(new_env_vars)

        return env

    def template_data(self):
        plugin_output = self.run_plugins()

        template_data = {}

        for k, v in plugin_output.items():
            if not isinstance(v, tuple) or len(v) != 2:
                msg = "Template plugin '%s' must return a tuple of length 2." % k
                raise dexy.exceptions.InternalDexyProblem(msg)
            template_data[k] = v[1]

        return template_data

    def process_text(self, input_text):
        template_data = self.template_data()
        return input_text % template_data

class JinjaFilter(TemplateFilter):
    """
    Runs the Jinja templating engine.
    """
    aliases = ['jinja']

    _settings = {
            'block-start-string' : ("Tag to indicate the start of a block.", "{%"),
            'block-end-string' : ("Tag to indicate the start of a block.", "%}"),
            'variable-start-string' : ("Tag to indicate the start of a variable.", "{{"),
            'variable-end-string' : ("Tag to indicate the start of a variable.", "}}"),
            'comment-start-string' : ("Tag to indicate the start of a comment.", "{#"),
            'comment-end-string' : ("Tag to indicate the start of a comment.", "#}"),
            'changetags' : ("Automatically change from { to < based tags for .tex and .wiki files.", True),
            'jinja-path' : ("List of additional directories to pass to jinja loader.", []),
            'jinja-extensions' : ("List of jinja extensions to activate.", ['jinja2.ext.do']),
            'workspace-includes' : [".jinja"],
            'assertion-passed-indicator' : (
                "Extra text to return with a passed assertion.",
                ""),
            'filters' : (
                "List of template plugins to make into jinja filters.",
                ['assertions', 'highlight', 'head', 'tail', 'rstcode', 'stripjavadochtml',
                    'replacejinjafilters', 'bs4']
                )
            }

    _not_jinja_settings = (
            'changetags',
            'jinja-path',
            'workspace-includes',
            'filters',
            'assertion-passed-indicator',
            'jinja-extensions'
            )

    TEX_TAGS = {
            'block_start_string': '<%',
            'block_end_string': '%>',
            'variable_start_string': '<<',
            'variable_end_string': '>>',
            'comment_start_string': '<#',
            'comment_end_string': '#>'
            }

    LYX_TAGS = {
            'block_start_string': '<%',
            'block_end_string': '%>',
            'variable_start_string': '<<',
            'variable_end_string': '>>',
            'comment_start_string': '<<#',
            'comment_end_string': '#>>'
            }

    def setup_jinja_env(self, loader=None):
        env_attrs = {}

        for k, v in self.setting_values().items():
            underscore_k = k.replace("-", "_")
            if k in self.__class__._settings and not k in self._not_jinja_settings:
                env_attrs[underscore_k] = v

        env_attrs['undefined'] = PassThroughWhitelistUndefined

        if self.ext in (".tex", ".wiki") and self.setting('changetags'):
            if 'lyxjinja' in self.doc.filter_aliases:
                tags = self.LYX_TAGS
            else:
                tags = self.TEX_TAGS

            self.log_debug("Changing tags to latex/wiki format: %s" % ' '.join(tags))

            for underscore_k, v in tags.items():
                hyphen_k = underscore_k.replace("_", "-")
                if env_attrs[underscore_k] == self.__class__._settings[hyphen_k][1]:
                    self.log_debug("setting %s to %s" % (underscore_k, v))
                    env_attrs[underscore_k] = v

        if loader:
            env_attrs['loader'] = loader

        extensions = []
        for ext in self.setting('jinja-extensions'):
            self.log_debug("attempting to activate %s" % ext)
            if ext.startswith("jinja2.ext"):
                ref = jinja2.ext.__dict__[ext.lstrip("jinja2.ext")]
                extensions.append(ref)
        env_attrs['extensions'] = extensions

        debug_attr_string = ", ".join("%s: %r" % (k, v) for k, v in env_attrs.items())
        self.log_debug("creating jinja2 environment with: %s" % debug_attr_string)
        return jinja2.Environment(**env_attrs)

    def handle_jinja_exception(self, e, input_text, template_data):
        result = []
        input_lines = input_text.splitlines()

        # Try to parse line number from stack trace...
        if isinstance(e, UndefinedError) or isinstance(e, TypeError):
            # try to get the line number
            m = re.search(r"File \"<template>\", line ([0-9]+), in top\-level template code", traceback.format_exc())
            if m:
                e.lineno = int(m.groups()[0])
            else:
                print((traceback.format_exc()))
                e.lineno = 0
                self.log_warn("unable to parse line number from traceback")

        args = {
                'error_type' : e.__class__.__name__,
                'key' : self.key,
                'lineno' : e.lineno,
                'message' : e.message,
                'name' : self.output_data.name,
                'workfile' : self.input_data.storage.data_file()
                }

        result.append("a %(error_type)s problem was detected: %(message)s" % args)

        if isinstance(e, UndefinedError):
            match_has_no_attribute = re.match("^'[\w\s\.]+' has no attribute '(.+)'$", e.message)
            match_is_undefined = re.match("^'([\w\s]+)' is undefined$", e.message)

            if match_has_no_attribute:
                undefined_object = match_has_no_attribute.groups()[0]
                match_lines = []
                for i, line in enumerate(input_lines):
                    if (".%s" % undefined_object in line) or ("'%s'" % undefined_object in line) or ("\"%s\"" % undefined_object in line):
                        result.append("line %04d: %s" % (i+1, line))
                        match_lines.append(i)
                if len(match_lines) == 0:
                    self.log_info("Tried to automatically find source of error: %s. Could not find match for '%s'" % (e.message, undefined_object))

            elif match_is_undefined:
                undefined_object = match_is_undefined.groups()[0]
                for i, line in enumerate(input_lines):
                    if undefined_object in line:
                        result.append("line %04d: %s" % (i+1, line))
            else:
                self.log_debug("Tried to automatically find where the error was in the template, but couldn't.")

        else:
            result.append("line %04d: %s" % (e.lineno, input_lines[e.lineno-1]))

        raise dexy.exceptions.UserFeedback("\n".join(result))

    def jinja_template_filters(self):
        filters = {}
        for alias in self.setting('filters'):
            self.log_debug("  creating filters from template plugin %s" % alias)
            template_plugin = TemplatePlugin.create_instance(alias)

            if not template_plugin.is_active():
                self.log_debug("    skipping %s - not active" % alias)
                continue
        
            methods = template_plugin.run()

            for k, v in methods.items():
                if not k in template_plugin.setting('no-jinja-filter'):
                    self.log_debug("    creating jinja filter for method %s" % k)
                    filters[k] = v[1]

        return filters

    def process(self):
        self.populate_workspace()

        wd = self.parent_work_dir()

        macro_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', 'macros'))
        dirs = ['.', wd, os.path.dirname(self.doc.name), macro_dir] + self.setting('jinja-path')
        self.log_debug("setting up jinja FileSystemLoader with dirs %s" % ", ".join(dirs))
        loader = FileSystemLoader(dirs)

        self.log_debug("setting up jinja environment")
        env = self.setup_jinja_env(loader=loader)
        self.log_debug("setting up jinja template filters")
        env.filters.update(self.jinja_template_filters())

        self.log_debug("initializing template")

        template_data = self.template_data()
        self.log_debug("jinja template data keys are %s" % ", ".join(sorted(template_data)))

        try:
            self.log_debug("about to create jinja template")
            template = env.get_template(self.work_input_filename())
            self.log_debug("about to process jinja template")
            template.stream(template_data).dump(self.output_filepath(), encoding="utf-8")
        except (TemplateSyntaxError, UndefinedError, TypeError) as e:
            self.log_debug("%s error while running jinja... %s" % (e.__class__.__name__, str(e)))
            try:
                self.log_debug("removing %s since jinja had an error" % self.output_filepath())
                os.remove(self.output_filepath())
            except os.error:
                pass
            self.handle_jinja_exception(e, str(self.input_data), template_data)
        except TemplateNotFound as e:
            msg = "Jinja couldn't find the template '%s', make sure this file is an input to %s" 
            msgargs = (e.message, self.doc.key)
            raise dexy.exceptions.UserFeedback(msg % msgargs)
        except Exception as e:
            try:
                self.log_debug("removing %s since jinja had an error" % self.output_filepath())
                os.remove(self.output_filepath())
            except os.error:
                pass
            self.log_debug(str(e))
            raise



from dexy.doc import Doc
#from dexy.utils import char_diff
from dexy.filters.templating import TemplateFilter
from dexy.filters.templating_plugins import TemplatePlugin
from tests.utils import wrap
from dexy.exceptions import UserFeedback

def test_jinja_invalid_attribute():
    def make_sections_doc(wrapper):
        return Doc("sections.txt",
                wrapper,
                [],
                contents = [{}, {"name" : "foo", "contents" : "This is foo."}]
                )

    with wrap() as wrapper:
        node = Doc("ok.txt|jinja",
                wrapper,
                [make_sections_doc(wrapper)],
                contents = """hello! foo contents are: {{ d['sections.txt'].foo }}"""
                )

        wrapper.run_docs(node)
        assert str(node.output_data()) == """hello! foo contents are: This is foo."""

    with wrap() as wrapper:
        node = Doc("broken.txt|jinja",
                wrapper,
                [make_sections_doc(wrapper)],
                contents = """There is no {{ d['sections.txt'].bar }}"""
                )
        try:
            wrapper.run_docs(node)
        except UserFeedback as e:
            assert str(e) == "No value for bar available in sections or metadata."

def test_jinja_pass_through():
    with wrap() as wrapper:
        with open("_template.html", "w") as f:
            f.write("{{ content }}")

        wrapper.reports = 'ws'
        contents = "{{ link(\"input.txt\") }}"
        doc = Doc("lines.html|jinja",
                    wrapper,
                    [
                        Doc("input.txt",
                            wrapper,
                            [],
                            contents = "nothing to see here"
                            )
                        ],
                    contents=contents,
                    apply_ws_to_content = True
                    )
        wrapper.run_docs(doc)
        assert str(doc.output_data()) == contents

        wrapper.report()

        with open("output-site/lines.html", 'r') as f:
            lines_html = f.read()
            assert lines_html == """<a href="input.txt">Input</a>"""

def test_jinja_pass_through_fails_if_not_whitelisted():
    with wrap() as wrapper:
        contents = "{{ linxxx('foo') }}"
        doc = Doc("lines.txt|jinja",
                    wrapper,
                    [],
                    contents=contents
                    )

        try:
            wrapper.run_docs(doc)
        except UserFeedback as e:
            assert "a UndefinedError problem" in str(e)
            assert "'linxxx' is undefined" in str(e)

def test_jinja_indent_function():
    with wrap() as wrapper:
        node = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("lines.txt",
                        wrapper,
                        [],
                        contents = "line one\nline two"
                        )
                    ],
                contents = "lines are:\n   {{ d['lines.txt'] | indent(3) }}"
                )
        wrapper.run_docs(node)
        assert str(node.output_data()) == """lines are:
   line one
   line two"""

def run_jinja_filter(contents):
    with wrap() as wrapper:
        doc = Doc("hello.txt|jinja",
                wrapper,
                [],
                contents = contents
                )
        wrapper.run_docs(doc)
        data = doc.output_data()
        data.data() # make sure is loaded
        return data

def test_jinja_filters_bs4():
    data = run_jinja_filter("{{ '<p>foo</p>' | prettify_html }}")
    assert str(data) == "<p>\n foo\n</p>"

def test_beautiful_soup_should_not_be_available_as_filter():
    try:
        run_jinja_filter("{{ 'foo' | BeautifulSoup }}")
        assert False
    except UserFeedback as e:
        assert "no filter named 'BeautifulSoup'" in str(e)

def test_jinja_filters_head():
    data = run_jinja_filter("{{ 'foo\nbar\nbaz' | head(1) }}")
    assert str(data) == "foo"
    data = run_jinja_filter("{{ 'foo\nbar\nbaz' | head(2) }}")
    assert str(data) == "foo\nbar"

def test_jinja_filters_tail():
    data = run_jinja_filter("{{ 'foo\nbar\nbaz' | tail(1) }}")
    assert str(data) == "baz"
    data = run_jinja_filter("{{ 'foo\nbar\nbaz' | tail(2) }}")
    assert str(data) == "bar\nbaz"

def test_jinja_filters_highlight():
    data = run_jinja_filter("{{ '<p>foo</p>' | highlight('html') }}")
    assert str(data) == """<div class="highlight"><pre><span></span><a name="l-1"></a><span class="p">&lt;</span><span class="nt">p</span><span class="p">&gt;</span>foo<span class="p">&lt;/</span><span class="nt">p</span><span class="p">&gt;</span>
</pre></div>\n"""
    
def test_jinja_filters_pygmentize():
    data = run_jinja_filter("{{ '<p>foo</p>' | pygmentize('html') }}")
    assert str(data)=="""<div class="highlight"><pre><span></span><a name="l-1"></a><span class="p">&lt;</span><span class="nt">p</span><span class="p">&gt;</span>foo<span class="p">&lt;/</span><span class="nt">p</span><span class="p">&gt;</span>
</pre></div>\n"""

def test_jinja_filters_combined():
    data = run_jinja_filter("{{ '<p>foo</p>' | prettify_html | highlight('html') }}")
    assert str(data) == """<div class="highlight"><pre><span></span><a name="l-1"></a><span class="p">&lt;</span><span class="nt">p</span><span class="p">&gt;</span>
<a name="l-2"></a> foo
<a name="l-3"></a><span class="p">&lt;/</span><span class="nt">p</span><span class="p">&gt;</span>
</pre></div>
"""

def test_jinja_kv():
    with wrap() as wrapper:
        node = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("blank.txt|keyvalueexample",
                        wrapper,
                        [],
                        contents = " ")
                    ],
                contents = """value of foo is '{{ d['blank.txt|keyvalueexample']['foo'] }}'"""
                )
        wrapper.run_docs(node)
        assert str(node.output_data()) == "value of foo is 'bar'"

def test_jinja_sectioned_invalid_section():
    with wrap() as wrapper:
        wrapper.debug = False
        doc = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("lines.txt|lines",
                        wrapper,
                        [],
                        contents = "line one\nline two"
                        )
                    ],
                contents = """first line is '{{ d['lines.txt|lines'][3] }}'"""
                )
        wrapper.run_docs(doc)
        assert wrapper.state == 'error'

def test_jinja_sectioned():
    with wrap() as wrapper:
        node = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("lines.txt|lines",
                        wrapper,
                        [],
                        contents = "line one\nline two")
                    ],
                contents = """first line is '{{ d['lines.txt|lines']['1'] }}'""")
        wrapper.run_docs(node)
        assert str(node.output_data()) == "first line is 'line one'"

def test_jinja_json():
    with wrap() as wrapper:
        node = Doc("hello.txt|jinja",
                wrapper,
                [
                    Doc("input.json",
                        wrapper, [],
                        contents = """{"foo":123}"""
                        )
                    ],
                contents = """foo is {{ d['input.json']['foo'] }}""")
        wrapper.run_docs(node)
        assert str(node.output_data()) == "foo is 123"

def test_jinja_undefined():
    with wrap() as wrapper:
        wrapper.debug = False
        node = Doc("template.txt|jinja",
                wrapper,
                [],
                contents = """{{ foo }}""")

        wrapper.run_docs(node)
        assert wrapper.state == 'error'

def test_jinja_syntax_error():
    with wrap() as wrapper:
        wrapper.debug = False
        node = Doc("template.txt|jinja",
                wrapper,
                [],
                contents = """{% < set foo = 'bar' -%}\nfoo is {{ foo }}\n"""
                )

        wrapper.run_docs(node)
        assert wrapper.state == 'error'

def test_jinja_filter_inputs():
    with wrap() as wrapper:
        node = Doc("template.txt|jinja",
                wrapper,
                [Doc("input.txt",
                    wrapper,
                    [],
                    contents = "I am the input.")
                ],
                contents = "The input is '{{ d['input.txt'] }}'")

        wrapper.run_docs(node)
        assert str(node.output_data()) == "The input is 'I am the input.'"

class TestSimple(TemplatePlugin):
    """
    test plugin
    """
    aliases = ['testtemplate']
    def run(self):
        return {'aaa' : ("docs", 1)}

class TestTemplateFilter(TemplateFilter):
    """
    test template
    """
    aliases = ['testtemplatefilter']

def test_template_filter_with_custom_filter_only():
    with wrap() as wrapper:
        node = Doc("hello.txt|testtemplatefilter",
                wrapper,
                [],
                contents = "aaa equals %(aaa)s",
                testtemplatefilter = { "plugins" : ["testtemplate"] }
                )

        wrapper.run_docs(node)
        assert node.output_data().as_text() == "aaa equals 1"
        plugins_used = node.filters[-1].template_plugins()
        assert len(plugins_used) == 1
        assert isinstance(plugins_used[0], TestSimple)

def test_jinja_filter():
    with wrap() as wrapper:
        node = Doc("template.txt|jinja",
                wrapper,
                [],
                contents = "1 + 1 is {{ 1+1 }}"
                )

        wrapper.run_docs(node)
        assert node.output_data().as_text() == "1 + 1 is 2"

def test_jinja_filter_tex_extension():
    with wrap() as wrapper:
        node = Doc("template.tex|jinja",
                wrapper,
                [],
                contents = "1 + 1 is << 1+1 >>")

        wrapper.run_docs(node)
        assert node.output_data().as_text() == "1 + 1 is 2"

def test_jinja_filter_custom_delims():
    with wrap() as wrapper:
        node = Doc("template.tex|jinja",
                wrapper,
                [],
                contents = "1 + 1 is %- 1+1 -%",
                jinja = {
                    "variable_start_string" : "%-",
                    "variable_end_string" : "-%"
                    }
                )

        wrapper.run_docs(node)
        assert node.output_data().as_text() == "1 + 1 is 2"

def test_jinja_filter_set_vars():
    with wrap() as wrapper:
        node = Doc("template.txt|jinja",
                wrapper,
                [],
                contents = """{% set foo = 'bar' -%}\nfoo is {{ foo }}\n"""
                )

        wrapper.run_docs(node)
        assert node.output_data().as_text() == "foo is bar"

def test_jinja_filter_using_inflection():
    with wrap() as wrapper:
        node = Doc("template.txt|jinja",
                wrapper,
                [],
                contents = """{{ humanize("abc_def") }}"""
                )

        wrapper.run_docs(node)
        assert node.output_data().as_text() == "Abc def"



# The easyhtml filter is defined in dexy.filters.fluid_html

from dexy.filter import DexyFilter
from pygments.formatters import LatexFormatter

class EasyLatex(DexyFilter):
    """
    Wraps your text in LaTeX article header/footer.
    Easy way to generate a document which can be compiled using LaTeX (includes
    Pygments syntax highlighting).
    """
    aliases = ['easylatex']
    _settings = {
            'input-extensions' : ['.tex'],
            'output-extensions' : ['.tex'],
            'documentclass' : ("The document class to generate.", "article"),
            'style' : ("The pygments style to use.", "default"),
            'title' : ("Title of article.", ""),
            'author' : ("Author of article.", ""),
            'date' : ("Date of article.", ""),
            'font' : ("The font size to use.", "11pt"),
            'papersize' : ("The document class to generate.", "a4paper"),
            "preamble" : ("Additional custom LaTeX content to include in header.", "")
            }

    def pygments_sty(self):
        formatter = LatexFormatter(style=self.setting('style'))
        return formatter.get_style_defs()

    def process_text(self, input_text):
        args = self.setting_values()
        args['input'] = input_text
        args['pygments'] = self.pygments_sty()

        if self.setting('title'):
            args['title'] = r"\title{%(title)s}" % args
            args['maketitle'] = r"\maketitle"
        else:
            args['title'] = ""
            args['maketitle'] = ""

        if self.setting('date'):
            args['date'] = r"\date{%(date)s}" % args

        if self.setting('author'):
            args['author'] = r"\author{%(author)s}" % args

        return self.template % args

    template = r"""\documentclass[%(font)s,%(papersize)s]{%(documentclass)s}
\usepackage{color}
\usepackage{fancyvrb}
%(pygments)s

%(preamble)s

%(title)s
%(author)s
%(date)s

\begin{document}

%(maketitle)s


%(input)s

\end{document}
"""



### "imports"
from dexy.plugin import Command
from dexy.version import DEXY_VERSION
import modargs.args as args
import cashew.exceptions
import dexy.exceptions
import dexy.wrapper
import inspect
import logging
import os
import sys
import warnings

### "load-plugins"
import dexy.load_plugins

### "import-all-commands"
from dexy.commands.info import links_command
#from dexy.commands.cite import cite_command
from dexy.commands.parsers import parsers_command
from dexy.commands.conf import conf_command
from dexy.commands.dirs import cleanup_command
from dexy.commands.dirs import reset_command
from dexy.commands.dirs import setup_command
from dexy.commands.env import env_command
from dexy.commands.env import datas_command
from dexy.commands.env import plugins_command
from dexy.commands.fcmds import fcmd_command
from dexy.commands.fcmds import fcmds_command
from dexy.commands.filters import filters_command
from dexy.commands.filters import filters_command as filter_command
from dexy.commands.grep import grep_command
from dexy.commands.info import info_command
from dexy.commands.it import dexy_command
from dexy.commands.it import it_command
from dexy.commands.it import targets_command
from dexy.commands.nodes import nodes_command
from dexy.commands.reporters import reporters_command
from dexy.commands.reporters import reporters_command as reports_command
from dexy.commands.serve import serve_command
from dexy.commands.templates import gen_command
from dexy.commands.templates import template_command
from dexy.commands.templates import templates_command

### "modargs-settings"
dexy_default_cmd = 'dexy'
dexy_cmd_mod = sys.modules[__name__]
prog = 'dexy'
### @end

def run():
    capture_warnings()
    parse_and_run_cmd(*resolve_argv())

def capture_warnings():
    """
    Capture deprecation messages and other irrelevant warnings in whatever way
    is appropriate to the dexy version.
    """
    if hasattr(logging, 'captureWarnings'):
        logging.captureWarnings(True)
    else:
        warnings.filterwarnings("ignore",category=Warning)

def resolve_argv():
    """
    Do some processing of the user-provided arguments in argv before they go to
    modargs so we can support commands defined in plugins.
    """
    only_one_arg = (len(sys.argv) == 1)
    second_arg_is_known_cmd = not only_one_arg and \
        sys.argv[1] in args.available_commands(dexy_cmd_mod)
    second_arg_is_option = not only_one_arg and \
        sys.argv[1].startswith("-")

    if only_one_arg or second_arg_is_known_cmd or second_arg_is_option:
        return sys.argv[1:], dexy_cmd_mod, dexy_default_cmd

    else:
        cmd, subcmd, cmd_mod = resolve_plugin_cmd(sys.argv[1])
        default_cmd = cmd.default_cmd or cmd.namespace
        return [subcmd] + sys.argv[2:], cmd_mod, default_cmd

def resolve_plugin_cmd(raw_command_name):
    """
    Take a command name like viewer:run and return the command method and
    module object.
    """
    if ":" in raw_command_name:
        alias, subcommand = raw_command_name.split(":")
    else:
        alias, subcommand = raw_command_name, ''

    try:
        cmd = Command.create_instance(alias)
    except cashew.exceptions.NoPlugin:
        msg = """No command '%s' available.
        Run `dexy help --all` to see list of available commands."""
        msgargs = (alias)
        sys.stderr.write(inspect.cleandoc(msg) % msgargs)
        sys.stderr.write(os.linesep)
        sys.exit(1)

    mod_name = cmd.__module__
    cmd_mod = args.load_module(mod_name)

    return cmd, subcommand, cmd_mod

def parse_and_run_cmd(argv, module, default_command):
    try:
        args.parse_and_run_command(argv, module, default_command)
    except (dexy.exceptions.UserFeedback, cashew.exceptions.UserFeedback) as e:
        # nice output for UserFeedback errors
        msg = """Oops, there's a problem running your command.
        Here is some more information:"""
        sys.stderr.write(inspect.cleandoc(msg))
        sys.stderr.write(os.linesep)

        err_msg = str(e)
        if err_msg:
            sys.stderr.write("'%s'" % str(e))
        else:
            sys.stderr.write("Sorry, can't get text of error message.")

        sys.stderr.write(os.linesep)

    except KeyboardInterrupt:
        sys.stderr.write("stopping...")
        sys.stderr.write(os.linesep)
        sys.exit(1)

def help_command(
        all=False, # List all available dexy commands (auto-generated).
        on=False # Get help on a particular dexy command.
    ):

    if all and not on:
        print("")
        args.help_command(prog, dexy_cmd_mod, dexy_default_cmd, on)
        print("")

    elif not on:
        print("""
        For help on the main `dexy` command, run `dexy help -on dexy`.

        The dexy tool includes several different commands:
          `dexy help --all` lists all available commands
          `dexy help --on XXX` provides help on a specific command
        
        Commands for running dexy:
          `dexy` runs dexy
          `dexy setup` makes directories dexy needs
          `dexy cleanup` removes directories dexy has created
          `dexy reset` empties and resets dexy's working directories
        
        Commands which print lists of dexy features:
          `dexy filters` filters like |jinja |py |javac
          `dexy reports` reporters like `output` and `run`
          `dexy nodes` node types and their document settings
          `dexy datas` data types and available methods
          `dexy env` elements available in document templates
        
        Commands which print information about your project:
          (you need to be in the project dir and have run dexy already)
          `dexy grep` search for documents and keys in documents
          `dexy info` list metadata about a particular document
          `dexy targets` list target names you can run
          `dexy links` list all ways to refer to documents and sections
        
        Other commands:
          `dexy serve` start a local static web server to view generated docs
          `dexy help` you're reading it
          `dexy version` print the version of dexy software which is installed
        """)

    else:
        try:
            args.help_command(prog, dexy_cmd_mod, dexy_default_cmd, on)

        except KeyError:
            sys.stderr.write("Could not find help on '%s'." % on)
            sys.stderr.write(os.linesep)
            sys.exit(1)


def version_command():
    """
    Print the version number of dexy.
    """
    print(f"{prog} version {DEXY_VERSION}")

def completion_command(on):
    args.completion_command(prog, dexy_cmd_mod, dexy_default_cmd, on)



import dexy.plugin

class WidgetBase(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Example of plugin.
    """
    _settings = {
            'foo' : ("Default value for foo", "bar"),
            'abc' : ("Default value for abc", 123)
            }

class Widget(WidgetBase):
    """
    Widget class.
    """
    aliases = ['widget']

class SubWidget(Widget):
    """
    Subwidget class.
    """
    aliases = ['sub']
    _settings = {
            'foo' : 'baz'
            }

class Fruit(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    '''fruit class'''
    aliases = ['fruit']
    _settings = {}

class Starch(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    '''starch class'''
    aliases = ['starch']
    _settings = {}
    _other_class_settings = {
            'fruit' : {
                    "color" : ("The color of the fruit", "red")
                }
            }

def test_plugin_meta():
    new_class = dexy.plugin.PluginMeta(
            "Foo",
            (dexy.plugin.Plugin,),
            {
                'aliases' : [],
                "__doc__" : 'help',
                '__metaclass__' : dexy.plugin.PluginMeta
                }
            )

    assert new_class.__name__ == 'Foo'
    assert new_class.__doc__ == 'help'
    assert new_class.aliases == []
    assert new_class.plugins == {}

def test_create_instance():
    widget = Widget.create_instance('widget')
    assert widget.setting('foo') == 'bar'

    sub = Widget.create_instance('sub')
    assert sub.setting('foo') == 'baz'
    assert sub.setting('abc') == 123
    assert sub.setting_values()['foo'] == 'baz'
    assert sub.setting_values()['abc'] == 123

def test_other_class_settings():
    fruit = Fruit()
    fruit.initialize_settings()
    assert fruit.setting('color') == 'red'



from dexy.utils import md5_hash
from dexy.utils import os_to_posix
import dexy.doc
import dexy.plugin
import fnmatch
import json
import re

class Node(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    base class for Nodes
    """
    _settings = {}
    aliases = ['node']
    state_transitions = (
            ('new', 'cached'),
            ('cached', 'consolidated'),
            ('new', 'uncached'),
            ('uncached', 'running'),
            ('running', 'ran'),
            )

    def __init__(self, pattern, wrapper, inputs=None, **kwargs):
        self.key = os_to_posix(pattern)
        self.wrapper = wrapper
        self.args = kwargs
        if inputs:
            self.inputs = list(inputs)
        else:
            self.inputs = []

        self.initialize_settings(**kwargs)

        self.start_time = 0
        self.finish_time = 0
        self.elapsed_time = 0

        self.runtime_args = {}
        self.children = []
        self.additional_docs = []

        self.hashid = md5_hash(self.key)

        self.state = 'new'

        # Class-specific setup.
        self.setup()

    def setup(self):
        pass
   
    def check_doc_changed(self):
        return False

    def __repr__(self):
        return "%s(%s)" % ( self.__class__.__name__, self.key)

    def transition(self, new_state):
        dexy.utils.transition(self, new_state)

    def update_all_settings(self, new_args):
        pass

    def add_runtime_args(self, args):
        self.update_all_settings(args)
        self.runtime_args.update(args)
        self.wrapper.batch.update_doc_info(self)

    def arg_value(self, key, default=None):
        return self.args.get(key, default) or self.args.get(key.replace("-", "_"), default)

    def walk_inputs(self):
        """
        Yield all direct inputs and their inputs.
        """
        children = []
        def walk(inputs):
            for inpt in inputs:
                children.append(inpt)
                walk(inpt.inputs + inpt.children)

        if self.inputs:
            walk(self.inputs)
        elif hasattr(self, 'parent'):
            children = self.parent.walk_inputs()

        return children

    def walk_input_docs(self):
        """
        Yield all direct inputs and their inputs, if they are of class 'doc'
        """
        for node in self.walk_inputs():
            if node.__class__.__name__ == 'Doc':
                yield node

    def log_debug(self, message):
        self.wrapper.log.debug("(state:%s) %s %s: %s" % (self.wrapper.state, self.hashid, self.key_with_class(), message))

    def log_info(self, message):
        self.wrapper.log.info("(state:%s) %s %s: %s" % (self.wrapper.state, self.hashid, self.key_with_class(), message))

    def log_warn(self, message):
        self.wrapper.log.warn("(state:%s) %s %s: %s" % (self.wrapper.state, self.hashid, self.key_with_class(), message))

    def key_with_class(self):
        return "%s:%s" % (self.__class__.aliases[0], self.key)

    def check_args_changed(self):
        """
        Checks if args have changed by comparing calculated hash against the
        archived calculated hash from last run.
        """
        saved_args = self.wrapper.saved_args.get(self.key_with_class())
        if not saved_args:
            self.log_debug("no saved args, will return True for args_changed")
            return True
        else:
            self.log_debug("    saved args '%s' (%s)" % (saved_args, saved_args.__class__))
            self.log_debug("    sorted args '%s' (%s)" % (self.sorted_arg_string(), self.sorted_arg_string().__class__))
            self.log_debug("  args unequal: %s" % (saved_args != self.sorted_arg_string()))
            return saved_args != self.sorted_arg_string()

    def sorted_args(self, skip=['contents']):
        """
        Returns a list of args in sorted order.
        """
        if not skip:
            skip = []

        sorted_args = []
        for k in sorted(self.args):
            if not k in skip:
                sorted_args.append((k, self.args[k]))
        return sorted_args

    def sorted_arg_string(self):
        """
        Returns a string representation of args in sorted order.
        """
        return str(json.dumps(self.sorted_args()))

    def additional_doc_info(self):
        additional_doc_info = []
        for doc in self.additional_docs:
            info = (doc.key, doc.hashid, doc.setting_values())
            additional_doc_info.append(info)
        return additional_doc_info

    def load_additional_docs(self, additional_doc_info):
        for doc_key, hashid, doc_settings in additional_doc_info:
            new_doc = dexy.doc.Doc(doc_key,
                    self.wrapper,
                    [],
                    **doc_settings
                    )
            new_doc.contents = None
            new_doc.args_changed = False
            new_doc.state = 'cached'
            assert new_doc.hashid == hashid

            new_doc.check_is_cached()
            new_doc.consolidate_cache_files()

            new_doc.initial_data.load_data()
            new_doc.output_data().load_data()
            self.add_additional_doc(new_doc)

    def add_additional_doc(self, doc):
        self.log_debug("adding additional doc '%s'" % doc.key)
        doc.created_by_doc = self
        self.children.append(doc)
        self.wrapper.add_node(doc)
        self.wrapper.batch.add_doc(doc)
        self.additional_docs.append(doc)

    def check_cache_elements_present(self):
        """
        Verify that all expected cache files are in fact present.
        """
        return True

    def input_nodes(self, with_parent_inputs = False):
        input_nodes = self.inputs + self.children
        if with_parent_inputs and hasattr(self, 'parent'):
            if not self in self.parent.inputs:
                input_nodes.extend(self.parent.inputs)
        return input_nodes

    def check_is_cached(self):
        if self.state == 'new':
            self.log_debug("checking if %s is changed" % self.key)

            any_inputs_not_cached = False
            for node in self.input_nodes(True):
                node.check_is_cached()
                if not node.state == 'cached':
                    self.log_debug("    input node %s is not cached" % node.key_with_class())
                    any_inputs_not_cached = True

            self.args_changed = self.check_args_changed()
            self.doc_changed = self.check_doc_changed()
            cache_elements_present = self.check_cache_elements_present()
                
            self.log_debug("  doc changed %s" % self.doc_changed)
            self.log_debug("  args changed %s" % self.args_changed)
            self.log_debug("  any inputs not cached %s" % any_inputs_not_cached)
            # log the 'not' so we can search for 'True' in logs to find uncached items
            self.log_debug("  cache elements missing %s" % (not cache_elements_present))

            is_cached = not self.doc_changed and not self.args_changed and not any_inputs_not_cached

            if is_cached and cache_elements_present:
                self.transition('cached')
            else:
                self.transition('uncached')

            # do housekeeping stuff we need to do for every node
            self.wrapper.add_node(self)
            self.wrapper.batch.add_doc(self)

    def load_runtime_info(self):
        pass

    def consolidate_cache_files(self):
        for node in self.input_nodes():
            node.consolidate_cache_files()

        if self.state == 'cached':
            self.transition('consolidated')

    def __lt__(self, other):
        return self.key < other.key

    def __iter__(self):
        def next_task():
            if self.state == 'uncached':
                self.transition('running')
                self.log_info("running...")
                yield self
                self.transition('ran')

            elif self.state in ('consolidated',):
                self.log_info("using cache for self and any children")

            elif self.state in ('ran',):
                self.log_info("already ran in this batch")

            elif self.state == 'running':
                raise dexy.exceptions.CircularDependency(self.key)

            else:
                raise dexy.exceptions.UnexpectedState("%s in %s" % (self.state, self.key))

        return next_task()

    def __call__(self, *args, **kw):
        for inpt in self.inputs:
            for task in inpt:
                task()
        self.wrapper.current_task = self
        self.run()
        self.wrapper.current_task = None

    def run(self):
        """
        Method which processes node's content if not cached, also responsible
        for calling child nodes.
        """
        for child in self.children:
            for task in child:
                task()

class BundleNode(Node):
    """
    Acts as a wrapper for other nodes.
    """
    aliases = ['bundle']

class ScriptNode(BundleNode):
    """
    Represents a bundle of nodes which need to run in order.

    If any of the bundle siblings change, the whole bundle should be re-run.
    """
    aliases = ['script']

    def check_doc_changed(self):
        return any(i.doc_changed for i in self.inputs)

    def setup(self):
        self.script_storage = {}

        siblings = []
        for doc in self.inputs:
            doc.parent = self
            doc.inputs = doc.inputs + siblings
            siblings.append(doc)

#        self.doc_changed = self.check_doc_changed()
#
#        for doc in self.inputs:
#            if not self.doc_changed:
#                assert not doc.doc_changed
#            doc.doc_changed = self.doc_changed

class PatternNode(Node):
    """
    Represents a file matching pattern.

    Creates child Doc objects for each file which matches the pattern.
    """
    aliases = ['pattern']

    def check_doc_changed(self):
        return any(child.doc_changed for child in self.children)

    def setup(self):
        file_pattern = self.key.split("|")[0]
        filter_aliases = self.key.split("|")[1:]

        for filepath, fileinfo in self.wrapper.filemap.items():
            if fnmatch.fnmatch(filepath, file_pattern):
                except_p = self.args.get('except')
                if except_p and re.search(except_p, filepath):
                    msg = "not creating child of patterndoc for file '%s' because it matches except '%s'"
                    msgargs = (filepath, except_p)
                    self.log_debug(msg % msgargs)
                else:
                    if len(filter_aliases) > 0:
                        doc_key = "%s|%s" % (filepath, "|".join(filter_aliases))
                    else:
                        doc_key = filepath

                    msg = "creating child of patterndoc %s: %s"
                    msgargs = (self.key, doc_key)
                    self.log_debug(msg % msgargs)
                    doc = dexy.doc.Doc(doc_key, self.wrapper, [], **self.args)
                    doc.parent = self
                    self.children.append(doc)
                    self.wrapper.add_node(doc)
                    self.wrapper.batch.add_doc(doc)



from dexy.filter import Filter
from tests.utils import runfilter
from nose.exc import SkipTest
from nose.tools import raises
from dexy.utils import s
from dexy.utils import split_path
from dexy.utils import iter_paths

def test_iter_path():
    full_path = "/foo/bar/baz"

    expected_paths = {
            0 : '/',
            1 : '/foo',
            2 : '/foo/bar',
            3 : '/foo/bar/baz'
            }

    for i, path in enumerate(iter_paths(full_path)):
        assert expected_paths[i] == path

def test_split_path():
    path = "foo/bar/baz"
    assert split_path(path) == ['foo', 'bar', 'baz']

def test_split_path_with_root():
    path = "/foo/bar/baz"
    assert split_path(path) == ['', 'foo', 'bar', 'baz']

def test_s():
    text = """This is some text
    which goes onto
    many lines and has
    indents at the start."""
    assert s(text) == 'This is some text which goes onto many lines and has indents at the start.'

class InactiveDexyFilter(Filter):
    """
    filter which is always inactive, for testing
    """
    aliases = ['inactive']

    def is_active(self):
        return False

@raises(SkipTest)
def test_inactive_filters_skip():
    with runfilter("inactive", "hello"):
        pass



from bs4 import BeautifulSoup
from dexy.filter import DexyFilter
from dexy.utils import chdir
import base64
import inflection
import mimetypes
import re
import urllib

class Customize(DexyFilter):
    """
    Add <script> tags or <link> tags to an HTML file's header.
    
    Uses BeautifulSoup.
    """
    aliases = ['customize']

    _settings = {
            'scripts' : ("Javascript files to add.", []),
            'stylesheets' : ("CSS files to add.", [])
            }

    def process_text(self, input_text):
        soup = BeautifulSoup(input_text)

        for js in self.setting('scripts'):
            js_tag = soup.new_tag("script", type="text/javascript", src=js)
            soup.head.append(js_tag)

        for css in self.setting('stylesheets'):
            css_tag = soup.new_tag("link", rel="stylesheet", type="text/css", href=css)
            soup.head.append(css_tag)

        return str(soup)

class InlineAssets(DexyFilter):
    """
    Imports any referenced images as data URIs.
    """
    aliases = ['inliner']

    _settings = {
            'html-parser' : ("Name of html parser BeautifulSoup should use.", 'html.parser'),
            'inline-images' : ("Whether to inline images using the data uri scheme.", True),
            'inline-styles' : ("Whether to embed referenced CSS in the page header.", True)
            }

    def inline_images(self, soup):
        for tag in soup.find_all("img"):
            path = tag.get('src')

            f = urllib.urlopen(path)
            data = f.read()
            f.close()

            mime, _ = mimetypes.guess_type(path)
            data64 = base64.encodestring(data)
            dataURI = 'data:%s;base64,%s' % (mime, data64)
            tag['src'] = dataURI

    def inline_styles(self, soup):
        for tag in soup.find_all("link"):
            path = tag.get('href')

            f = urllib.urlopen(path)
            data = f.read()
            f.close()

            style = soup.new_tag('style')
            style.string = data

            tag.replace_with(style)

    def process(self):
        soup = BeautifulSoup(str(self.input_data), self.setting('html-parser'))
        self.populate_workspace()

        with chdir(self.parent_work_dir()):
            if self.setting('inline-images'):
                self.inline_images(soup)
    
            if self.setting('inline-styles'):
                self.inline_styles(soup)
    
        self.output_data.set_data(str(soup))

class SoupSections(DexyFilter):
    """
    Split a HTML file into nested sections based on header tags.
    """
    aliases = ['soups']

    _settings = {
            'data-type' : 'sectioned',
            'html-parser' : ("Name of html parser BeautifulSoup should use.", 'html.parser'),
            'initial-section-name' : ("Name to use for the initial section which currently holds all the contents.", "Actual Document Contents"),
            }

    def append_current_section(self):
        section_dict = {
                "name" : self.current_section_name,
                "contents" : self.current_section_text,
                "level" : self.current_section_level,
                "id" : self.current_section_anchor
                }
        self.output_data._data.append(section_dict)

    def process(self):
        soup = BeautifulSoup(str(self.input_data), self.setting('html-parser'))

        for tag in soup.find_all(re.compile("^h[0-6]")):
            name = tag.text
            m = re.match("^h([0-6])$", tag.name)

            if not 'id' in tag.attrs:
                tag.attrs['id'] = inflection.parameterize(name)

            self.current_section_anchor = tag.attrs['id']
            self.current_section_text = None
            self.current_section_name = name
            self.current_section_level = int(m.groups()[0])

            self.append_current_section()

        self.current_section_text = str(soup)
        self.current_section_name = self.setting('initial-section-name')
        self.current_section_level = 1
        self.current_section_anchor = None

        self.append_current_section()

        self.output_data.save()



from dexy.filter import DexyFilter
import tarfile
import zipfile
import os

class UnprocessedDirectoryArchiveFilter(DexyFilter):
    """
    Create a .tgz archive containing the unprocessed files in a directory.
    """
    aliases = ['tgzdir']
    _settings = {
            'output' : True,
            'output-extensions' : ['.tgz'],
            'dir' : ("Directory in which to output the archive.", '')
            }

    def process(self):
        parent_dir = self.output_data.parent_dir()
        subdir = self.setting('dir')
        dir_to_archive = os.path.join(parent_dir, subdir)
        af = self.output_filepath()
        tar = tarfile.open(af, mode="w:gz")
        for fn in os.listdir(dir_to_archive):
            fp = os.path.join(dir_to_archive, fn)
            self.log_debug("Adding file %s to archive %s" % (fp, af))
            tar.add(fp, arcname=os.path.join(subdir, fn))
        tar.close()

class ArchiveFilter(DexyFilter):
    """
    Creates a .tgz archive of all input documents.

    The use-short-names option will store documents under their short
    (canonical) filenames.
    """
    aliases = ['archive', 'tgz']
    _settings = {
            'output' : True,
            'output-extensions' : ['.tgz'],
            'use-short-names' : ("Whether to use short, potentially non-unique names within the archive.", False),
            }

    def open_archive(self):
        self.archive = tarfile.open(self.output_filepath(), mode="w:gz")

    def add_to_archive(self, filepath, archivename):
        self.archive.add(filepath, arcname=archivename)

    def process(self):
        self.open_archive()

        # Place files in the archive within a directory with the same name as the archive.
        dirname = self.output_data.baserootname()

        # Figure out whether to use short names or longer, unambiguous names.
        use_short_names = self.setting("use-short-names")

        for doc in self.doc.walk_input_docs():
            if not doc.output_data().is_cached():
                raise Exception("File not on disk.")

            # Determine what this file's name within the archive should be.
            if use_short_names:
                arcname = doc.output_data().name
            else:
                arcname = doc.output_data().long_name()
            arcname = os.path.join(self.input_data.relative_path_to(arcname))
            arcname = os.path.join(dirname, arcname)

            # Add file to archive
            self.add_to_archive(doc.output_data().storage.data_file(), arcname)

        # Save the archive
        self.archive.close()

class ZipArchiveFilter(ArchiveFilter):
    """
    Creates a .zip archive of all input documents.

    The use-short-names option will store documents under their short
    (canonical) filenames.
    """
    aliases = ['zip']
    _settings = {
            'output-extensions' : ['.zip']
            }

    def open_archive(self):
        self.archive = zipfile.ZipFile(self.output_filepath(), mode="w")

    def add_to_archive(self, filepath, archivename):
        self.archive.write(filepath, arcname=archivename)



from dexy.exceptions import UserFeedback
from dexy.filters.git import repo_from_path
from dexy.filters.git import repo_from_url
from dexy.filters.git import generate_commit_info
from tests.utils import assert_in_output
from tests.utils import runfilter
from tests.utils import tempdir
from nose.exc import SkipTest
import os
import json

REMOTE_REPO_HTTPS = "https://github.com/ananelson/dexy-templates"
PATH_TO_LOCAL_REPO = os.path.expanduser("~/dev/testrepo")
# TODO use subprocess to check out a repo to a temp dir, or have a repo in data
# dir, or use [gasp] submodules.

try:
    import pygit2
    import urllib

    no_local_repo = not os.path.exists(PATH_TO_LOCAL_REPO)

    try:
        urllib.urlopen("http://google.com")
        no_internet = False
    except IOError:
        no_internet = True

    if no_local_repo:
        SKIP = (True, "No local repo at %s." % PATH_TO_LOCAL_REPO)
    elif no_internet:
        SKIP = (True, "Internet not available.")
    else:
        SKIP = (False, None)

except ImportError:
    SKIP = (True, "pygit2 not installed")

def skip():
    if SKIP[0]:
        raise SkipTest(SKIP[1])

skip()

def test_run_gitrepo():
    with runfilter("repo", REMOTE_REPO_HTTPS) as doc:
        assert len(doc.wrapper.nodes) > 20

def test_generate_commit_info():
    repo, remote = repo_from_url(REMOTE_REPO_HTTPS)

    refs = repo.listall_references()
    ref = repo.lookup_reference(refs[0])
    commit = repo[ref.target]
    commit_info = generate_commit_info(commit)

    assert commit_info['author-name'] == "Ana Nelson"
    assert commit_info['author-email'] == "ana@ananelson.com"

def test_git_commit():
    with runfilter("gitcommit", REMOTE_REPO_HTTPS) as doc:
        output = doc.output_data()
        patches = json.loads(output['patches'])
        assert output['author-name'] == "Ana Nelson"
        assert output['author-email'] == "ana@ananelson.com"
        #assert output['message'] == "Add README file."
        #assert output['hex'] == "2f15837e64a70e4d34b924f6f8c371a266d16845"

def test_git_log():
    assert_in_output("gitlog", PATH_TO_LOCAL_REPO,
            "Add README file.")

def test_git_log_remote():
    assert_in_output("gitlog", REMOTE_REPO_HTTPS,
            "Rename")

def test_repo_from_url():
    repo, remote = repo_from_url(REMOTE_REPO_HTTPS)
    assert remote.name == 'origin'
    assert remote.url == REMOTE_REPO_HTTPS

def test_repo_from_path():
    repo, remote = repo_from_path(PATH_TO_LOCAL_REPO)
    assert ".git" in repo.path
    #assert isinstance(repo.head, pygit2.Object)
    # assert "README" in repo.head.message

def test_repo_from_invalid_path():
    with tempdir():
        try:
            repo, remote = repo_from_path(".")
            assert False
        except UserFeedback as e:
            assert "no git repository was found at '.'" in str(e)

def test_run_git():
    with runfilter("git", PATH_TO_LOCAL_REPO) as doc:
        doc.output_data()

def test_run_git_remote():
    with runfilter("git", REMOTE_REPO_HTTPS) as doc:
        doc.output_data()



from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.filters.process import SubprocessExtToFormatFilter
import os

class Asciidoctor(SubprocessExtToFormatFilter):
    """
    Runs `asciidoctor`.
    """
    aliases = ['asciidoctor']
    _settings = {
            'tags' : ['asciidoc', 'html'],
            'examples' : ['asciidoctor'],
            'output' : True,
            'version-command' : "asciidoctor --version",
            'executable' : 'asciidoctor',
            'input-extensions' : ['.*'],
            'output-extensions': ['.html', '.xml', '.tex'],
            'stylesheet' : ("Custom asciidoctor stylesheet to use.", None),
            'format-specifier': '-b ',
            'backend' : ("Asciidoctor backend to use (optional, only to override default).", None),
            'ext-to-format' : {
                '.html' : 'html5',
                '.xml': 'docbook5',
                '.tex' : 'latex'
                },
            'command-string': '%(prog)s %(format)s %(args)s %(ss)s -o %(output_file)s %(script_file)s'
            }

    def command_string_args(self):
        args = super(Asciidoctor, self).command_string_args()

        stylesheet = self.setting('stylesheet')
        if stylesheet is not None:
            stylesdir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'asciidoctor'))

            if not os.path.exists(stylesdir):
                msg = "Asciidoctor stylesheet directory not found at '%s'"
                raise InternalDexyProblem(msg % stylesdir)

            args['ss'] = "-a stylesheet=%s -a stylesdir=%s" % (stylesheet, stylesdir)

            if not os.path.exists(os.path.join(stylesdir, stylesheet)):
                msg = "No stylesheet file named '%s' was found in directory '%s'. Files found: %s"
                stylesheets = os.listdir(stylesdir)
                raise UserFeedback(msg % (stylesheet, stylesdir, ", ".join(stylesheets)))

        else:
            args['ss'] = ''

        backend = self.setting('backend')
        if backend is not None:
            format_specifier = self.setting('format-specifier')
            args['format'] = "%s%s" % (format_specifier, backend)

        return args



from dexy.utils import s
import dexy.plugin
import dexy.utils
import dexy.wrapper
import os
import shutil
import sys

class Template(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Parent class for templates.
    """
    _settings = {
            'contents-dir' : ("Directory containing contents of template.", None),
            'nodoc' : ("Whether to not document this template.", False),
            'copy-output-dir' : ("", False)
            }
    aliases = []
    filters_used = []

    def __init__(self):
        self.initialize_settings()

    def template_source_dir(self):
        if self.safe_setting('install-dir'):
            template_install_dir = self.setting('install-dir')
        else:
            template_install_dir = os.path.dirname(sys.modules[self.__module__].__file__)

        if self.setting('contents-dir'):
            contents_dirname = self.setting('contents-dir')
        else:
            if self.__class__.aliases:
                canonical_alias = self.__class__.aliases[0]
            else:
                canonical_alias = self.alias

            # default is to have contents in directory with same name as alias followed by "-template"
            contents_dirname = "%s-template" % canonical_alias

        return os.path.join(template_install_dir, contents_dirname)

    def generate(self, directory, **kwargs):
        """
        Generates the template, making a copy of the template's files in
        the specified directory. Does not run dexy.
        """
        if dexy.utils.file_exists(directory):
            msg = "directory '%s' already exists, aborting" % directory
            raise dexy.exceptions.UserFeedback(msg)

        # copy template files
        source = self.template_source_dir()
        shutil.copytree(source, directory)

        # remove template documentation unless 'meta' is specified
        if not kwargs.get('meta'):
            dexy_rst = os.path.join(directory, 'dexy.rst')
            if dexy.utils.file_exists(dexy_rst):
                os.remove(dexy_rst)

    def dexy(self, meta=True, additional_doc_keys=None):
        """
        Run dexy on this template's files in a temporary directory.

        Yields the batch object for the dexy run, so we can call methods on it
        while still in the tempdir.
        """
        meta_doc_keys = [
                ".*",
                "dexy.yaml|idio|t",
                "dexy.rst|idio|t",
                "dexy.rst|jinja|rst2html",
                "dexy.rst|jinja|rst2man"
                ]

        with dexy.utils.tempdir():
            # Copy template files to directory 'ex'
            self.generate("ex", meta=meta)

            # Run dexy in directory 'ex'
            os.chdir("ex")
            wrapper = dexy.wrapper.Wrapper()
            wrapper.create_dexy_dirs()
            wrapper = dexy.wrapper.Wrapper(log_level='DEBUG')
            wrapper.to_valid()
            wrapper.nodes = {}
            wrapper.roots = []
            wrapper.batch = dexy.batch.Batch(wrapper)
            wrapper.filemap = wrapper.map_files()

            ast = wrapper.parse_configs()
            if additional_doc_keys:
                for doc_key in additional_doc_keys:
                    ast.add_node(doc_key)

            if meta and dexy.utils.file_exists('dexy.rst'):
                for doc_key in meta_doc_keys:
                    ast.add_node(doc_key)
                    if 'jinja' in doc_key:
                        for task in list(ast.lookup_table.keys()):
                            if not task in meta_doc_keys:
                                ast.add_dependency(doc_key, task)

            ast.walk()

            wrapper.transition('walked')
            wrapper.to_checked()

            try:
                wrapper.run()
            except (Exception, SystemExit,) as e:
                error = str(e)
                template_dir = os.path.abspath(".")
                msg = "%s\npushd %s" % (error, template_dir)
                raise dexy.exceptions.TemplateException(msg)

            if wrapper.state == 'error':
                template_dir = os.path.abspath(".")
                msg = "pushd %s" % (template_dir)
                raise dexy.exceptions.TemplateException(msg)

            yield(wrapper)

    def validate(self):
        """
        Runs dexy and validates filter list.
        """
        for wrapper in self.dexy(False):
            filters_used = wrapper.batch.filters_used

            for f in self.__class__.filters_used:
                msg = "filter %s not used by %s" % (f, self.__class__.__name__)
                assert f in filters_used, msg
    
            for f in filters_used:
                if not f.startswith('-') and not f in self.__class__.filters_used:
                    msg = s("""filter %(filter)s used by %(template)s
                            but not listed in klass.filters_used,
                            adjust list to: filters_used = [%(list)s]""")
                    msgargs = {
                            'filter' : f,
                            'template' : self.__class__.__name__,
                            'list'  : ", ".join("'%s'" % f for f in filters_used)
                            }
                    print((msg % msgargs))

            return wrapper.state == 'ran'
